# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""MySQL backend"""

import re

from cf.db import objects
from cf.db.backends import Generic, DEFAULT_OPTIONS

P_ERROR_MESSAGES = {
    1064: re.compile(r"near '(?P<pattern>.*)' at line (?P<lineno>\d+)$",
                     re.MULTILINE|re.UNICODE),
    1051: re.compile(r"Unknown table '(?P<pattern>.*)'$"),
    1146: re.compile(r"Table '.*\.(?P<pattern>.*)' doesn't exist$"),
    1054: re.compile(r"^Unknown column '(?P<pattern>.*)' in"),
}


class MySQL(Generic):

    drivername = 'mysql'

    @classmethod
    def dbapi(cls):
        import MySQLdb
        return MySQLdb

    @classmethod
    def get_native_shell_command(cls, url):
        args = []
        if url.user:
            args.extend(['-u', url.user])
        if url.password:
            args.extend(['--password=%s' % url.password])
        if url.host:
            args.extend(['-h', url.host])
        if url.port:
            args.extend(['-P', url.port])
        if url.database:
            args.extend(['-D', url.database])
        return 'mysql', args

    @classmethod
    def get_connect_params(cls, url):
        opts = url.get_dict()
        if 'database' in opts:
            opts['db'] = opts['database']
            del opts['database']
        if 'username' in opts:
            opts['user'] = opts['username']
            del opts['username']
        if 'password' in opts:
            opts['passwd'] = opts['password']
            del opts['password']
        opts.update(url.query)
        return tuple(), opts

    @classmethod
    def get_options(cls):
        return DEFAULT_OPTIONS

    @classmethod
    def create_url(cls, options):
        url = super(MySQL, cls).create_url(options)
        url.query = {'charset': 'utf8'}
        return url

    @classmethod
    def should_close(cls, error):
        if not isinstance(error, cls.dbapi().OperationalError):
            return False
        err_code, msg = error.args
        if err_code >= 2000:
            return True
        return False

    @classmethod
    def get_error_position(cls, query, error):
        err_code, msg = error.args
        p = P_ERROR_MESSAGES.get(err_code)
        if p is None:
            return None
        m = p.search(msg)
        if m is None:
            return None
        data = m.groupdict()
        line = data.get('lineno')
        pattern = data.get('pattern')
        if line is None and pattern is None:
            return None
        elif pattern is None:
            return int(line), 1
        if line is None:
            for idx, l in enumerate(query.statement.splitlines()):
                if pattern in l:
                    return idx+1, l.index(pattern)+1
        else:
            l = query.statement.splitlines()[int(line)-1]
            return int(line), l.index(pattern)+1

    def get_server_info(self, connection):
        return 'MySQL %s' % connection.connection.get_server_info()

    def initialize(self, meta, connection):
        schemata = objects.Schemata(meta)
        users = objects.Users(meta)
        meta.set_object(schemata)
        meta.set_object(users)
        for item in self._query(connection, INITIAL_SQL):
            if item['type'] == 'schema':
                schema = objects.Schema(meta, parent=schemata,
                                        oid=item['id'], name=item['name'],
                                        comment=item['description'])
                meta.set_object(schema)
                tables = objects.Tables(meta, parent=schema)
                meta.set_object(tables)
                schema.tables = tables
                views = objects.Views(meta, parent=schema)
                meta.set_object(views)
                schema.views = views

            if item['type'] in ('BASE TABLE', 'VIEW', 'SYSTEM VIEW'):
                schema = meta.find_exact(cls=objects.Schema,
                                         oid=item['parent'])
                if item['type'] == 'BASE TABLE':
                    cls = objects.Table
                    parent = meta.find_exact(cls=objects.Tables, parent=schema)
                else:
                    cls = objects.View
                    parent = meta.find_exact(cls=objects.Views, parent=schema)
                obj = cls(meta, parent=parent, oid=item['id'],
                          name=item['name'], comment=item['description'],
                          schema=schema)
                meta.set_object(obj)

            if item['type'] == 'column':
                parent = meta.find_exact(oid=item['parent'])
                col = objects.Column(meta, parent=parent.columns,
                                     oid=item['id'],
                                     name=item['name'],
                                     comment=item['description'])
                meta.set_object(col)

    def _query(self, connection, sql):
        return connection.execute_raw(sql)

    def refresh(self, obj, meta, connection):
        if obj.typeid == 'users':
            self._refresh_users(obj, meta, connection)
        elif obj.typeid == 'constraints':
            self._refresh_constraints(obj, meta, connection)

    def _refresh_constraints(self, coll, meta, connection):
        sql = ("select lower(concat(table_schema, '.', table_name)) as parent,"
               "lower(concat(table_schema, '.', table_name, '.',"
               " constraint_name)) as id,"
               " case when constraint_type = 'PRIMARY KEY'"
               " then constraint_type"
               " else concat(constraint_type, ' (', constraint_name, ')')"
               " end as name"
               " from information_schema.table_constraints"
               " where lower(table_name) = '%s'"
               "  and lower(table_schema) = '%s'"
               % (coll.parent.name.lower(),
                  coll.parent.parent.parent.name.lower()))
        for item in self._query(connection, sql):
            con = meta.find_exact(oid=item['id'],
                                  parent=coll.parent.constraints)
            if con is None:
                con = objects.Constraint(meta, parent=coll.parent.constraints,
                                         oid=item['id'])
                meta.set_object(con)
            con.name = item['name']

    def _refresh_users(self, coll, meta, connection):
        sql = "select concat(user, '@', host) as name from mysql.user"
        for item in self._query(connection, sql):
            user = meta.find_exact(parent=coll, name=item['name'])
            if user is None:
                user = objects.User(meta, parent=coll, name=item['name'])
                user.props.has_children = False
                meta.set_object(user)


DRIVER = MySQL

INITIAL_SQL = """select * from (
select lower(schema_name) as id, schema_name as name,
null as description, 'schema' as type, null as parent, 0 as pos
from information_schema.schemata
union
select lower(concat(table_schema, '.', table_name)) as id,
table_name as name, table_comment as description, table_type as type,
lower(table_schema) as parent, 1 as pos
from information_schema.tables
union
select lower(concat(table_schema, '.', table_name, '.', column_name)) as id,
column_name as name, column_comment as description, 'column' as type,
lower(concat(table_schema, '.', table_name)) as parent, 3 as pos
from information_schema.columns
) x order by pos asc
"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Database backends."""

from gettext import gettext as _

from cf.db import DIALECTS
from cf.db import objects
from cf.db import TRANSACTION_IDLE
from cf.db import TRANSACTION_COMMIT_ENABLED
from cf.db import TRANSACTION_ROLLBACK_ENABLED
from cf.db.url import make_url, URL


class GUIOption(object):

    WIDGET_PASSWORD = 'password'
    WIDGET_PORT = 'port'
    WIDGET_CHECKBOX = 'checkbox'
    WIDGET_FILECHOOSER = 'filechooser'
    WIDGET_COMBO = 'combo'

    def __init__(self, key, label, widget=None, required=False,
                 choices=None, tooltip=None):
        self.key = key
        self.label = label
        self.widget = widget
        self.required = required
        self.choices = choices
        self.tooltip = tooltip


DEFAULT_OPTIONS = (
    GUIOption('host', _(u'Host')),
    GUIOption('port', _(u'Port'),
              widget=GUIOption.WIDGET_PORT),
    GUIOption('database', _(u'Database')),
    GUIOption('username', _(u'Username')),
    GUIOption('password', _(u'Password'),
              widget=GUIOption.WIDGET_PASSWORD),
    )


class Generic(object):

    drivername = None

    @classmethod
    def get_options(cls):
        """Return a list of :class:`GUIOption` instances.

        Backend implementations can overwrite this method to provide
        appropriate options. The default implementation returns a single
        option to give the URL directly.
        """
        return (GUIOption('url', _(u'URL'), required=True),)

    @classmethod
    def create_url(cls, options):
        """Create an URL from options.

        :param options: A dictionary mapping :class:`GUIOption` keys with
            values from UI.

        :returns: An :class:`URL` instance.
        """
        url = URL(cls.drivername)
        attr_names = ('username', 'password', 'host', 'port', 'database')
        backend_keys = [opt.key for opt in cls.get_options()]
        for key in options:
            if key not in backend_keys:
                continue
            elif key in attr_names and options.get(key, None) is not None:
                setattr(url, key, options[key])
            elif options.get(key, None) is not None:
                url.query[key] = options[key]
        return url

    @classmethod
    def get_native_shell_command(cls, url):
        """Return command and arguments to connect to native db shells.

        The default implementation returns ``None`` which disabled the
        native shell for this backend. If a native shell is available for
        this backend the method must return a 2-tuple (command, argv).
        """
        return None

    @classmethod
    def get_connect_params(cls, url):
        """Convert URL to connection arguments and keyword arguments."""
        return tuple(), url.get_dict()

    @classmethod
    def dbapi(cls):
        """Returns the DB-API2 module."""
        msg = 'Driver.dbapi() not implemented [%s]' % cls.drivername
        raise NotImplementedError(msg)

    @classmethod
    def should_close(cls, error):
        """Return True if the connection should be closed after this error.

        The default implementation returns False.
        """
        return False

    @classmethod
    def get_error_position(cls, query, error):
        """Return a 2-tuple (line, offset) for a given error.

        The default implementation return None to indicate that the backend
        isn't capable of getting the position in the SQL statement where
        the error happend.

        query is a Query instance. error is an instance of
        dbapi.ProgrammingError or dbapi.OperationalError.
        """
        return None

    def get_connection(self, url):
        """Returns a DB-API2 connection."""
        args, kwds = self.get_connect_params(url)
        conn = self.dbapi().connect(*args, **kwds)
        return conn

    def prepare_connection(self, dbapi_connection):
        """Prepares the DB-API2 connection.

        This method can be overwritten by backend implementations and is
        called after the DB-API2 connection is opened. The default
        implementation does nothing.
        """
        pass

    def get_server_info(self, connection):
        """Return human-readable server version info."""
        return DIALECTS[self.drivername]['name']

    def prepare_statement(self, sql):
        """Prepare statement before executing."""
        return sql

    def initialize(self, meta, connection):
        """Initialize basic meta data.

        Backend implentations should overwrite this method to provide
        initial meta information about database objects (e.g. table,
        view and column names). Additional information may be fetched during
        runtime using the :meth:`refresh` and :meth:`get_children` methods.
        The meta information initialized here is mainly used for
        auto-completion.

        The default implementation does nothing.
        """
        pass

    def refresh(self, parent, meta, connection):
        """Refresh child objects for parent."""
        pass

    def get_transaction_state(self, connection):
        """Determine transaction state for connection.

        This method can be overwritten by backend implementations to
        find a more precise answer.
        """
        return TRANSACTION_IDLE

    def begin(self, connection):
        """Begin transaction. Return True if handled."""
        connection.execute('BEGIN')
        return True

    def commit(self, connection):
        """Commit transaction."""
        connection.execute('COMMIT')

    def rollback(self, connection):
        """Rollback transaction."""
        connection.execute('ROLLBACK')

    def get_explain_statement(self, statement):
        """Return a SQL for executing EXPLAIN or None."""
        return "EXPLAIN %s" % statement



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""PostgreSQL backend."""

import re
from gettext import gettext as _

from cf.db import TRANSACTION_IDLE
from cf.db import TRANSACTION_COMMIT_ENABLED
from cf.db import TRANSACTION_ROLLBACK_ENABLED
from cf.db.backends import Generic, DEFAULT_OPTIONS, GUIOption
from cf.db import objects


P_ERRORLINE = re.compile(r'^(?P<unused>[A-Z]+ )(?P<lineno>\d+): (...)?(?P<pattern>.*)(...)?$', re.UNICODE)


class Postgres(Generic):

    drivername = 'postgres'

    @classmethod
    def get_options(cls):
        return DEFAULT_OPTIONS + (
            GUIOption('ssl_mode', _(u'SSL Mode'),
                      widget=GUIOption.WIDGET_COMBO,
                      choices=(
                          (None, None),
                          ('disable', _(u'No SSL (disable)')),
                          ('allow',
                           _(u'First try non SSL, then try SSL (allow)')),
                          ('prefer',
                           _(u'First try SSL, then non-SSL (prefer)')),
                          ('require',
                           _(u'Only try SSL connection (require)')),
                          )),
            )

    @classmethod
    def create_url(cls, options):
        url = super(Postgres, cls).create_url(options)
        if 'ssl_mode' in options and options['ssl_mode'] is not None:
            url.query = {'sslmode': options['ssl_mode']}
        return url

    @classmethod
    def get_native_shell_command(cls, url):
        args = []
        if url.username:
            args.extend(['-U', url.username])
        if url.database:
            args.extend(['-d', url.database])
        if url.host:
            args.extend(['-h', url.host])
        if url.port:
            args.extend(['-p', url.port])
        return 'psql', args

    @classmethod
    def dbapi(cls):
        import psycopg2
        return psycopg2

    @classmethod
    def get_connect_params(cls, url):
        opts = url.get_dict()
        if 'username' in opts:
            opts['user'] = opts['username']
            del opts['username']
        return tuple(), opts

    @classmethod
    def get_error_position(cls, query, error):
        if not isinstance(error, cls.dbapi().ProgrammingError):
            return None
        line_no = offset = skip_offset = pattern = None
        for line in error.pgerror.splitlines():
            # If the line number was found, the next line contains an
            # offset indicator.
            if line_no:
                offset_ws, unused = line[skip_offset:].split('^', 1)
                sql_line = query.statement.splitlines()[line_no-1]
                offset = len(offset_ws)+sql_line.index(pattern)-2
                break
            m = P_ERRORLINE.match(line)
            if m and line_no is None:
                line_no = int(m.groupdict()['lineno'])
                pattern = m.groupdict()['pattern']
                if pattern.endswith('...'):
                    pattern = pattern[:-3]
                skip_offset = len('%s%s: ' % (m.groupdict()['unused'],
                                              line_no))
        if line_no:
            return (line_no, offset)
        return None

    def prepare_connection(self, connection):
        import psycopg2.extensions as pe
        level = pe.ISOLATION_LEVEL_AUTOCOMMIT
        connection.connection.set_isolation_level(level)

    def get_server_info(self, connection):
        return connection.execute('select version()')[0][0]

    def _query(self, connection, sql):
        return connection.execute_raw(sql)

    def _get_search_path(self, connection):
        search_path = self._query(connection,
                                  'show search_path')[0][0].split(',')
        if '"$user"' in search_path:
            idx = search_path.index('"$user"')
            username = self._query(connection,
                                   'select current_user')[0][0]
            search_path[idx] = username
        return search_path

    def initialize(self, meta, connection):
        schemata = objects.Schemata(meta)
        users = objects.Users(meta)
        languages = objects.Languages(meta)
        meta.set_object(schemata)
        meta.set_object(users)
        meta.set_object(languages)
        search_path = self._get_search_path(connection)
        for item in self._query(connection, PG_INITIAL_SQL):
            if item['nspoid'] is not None:
                schema = meta.find_exact(oid=item['nspoid'],
                                              cls=objects.Schema,
                                              parent=schemata)
                if schema is None:
                    is_default = item['nspname'] in search_path
                    schema = objects.Schema(meta,
                                            oid=item['nspoid'],
                                            name=item['nspname'],
                                            comment=item['description'],
                                            parent=schemata,
                                            is_default=is_default)
                    meta.set_object(schema)
                    funcs = objects.Functions(meta, parent=schema)
                    meta.set_object(funcs)
            else:
                schema = None
            if item['objtype'] in ('table', 'view', 'sequence'):
                if item['objtype'] == 'table':
                    cls = objects.Table
                    coll_cls = objects.Tables
                elif item['objtype'] == 'sequence':
                    cls = objects.Sequence
                    coll_cls = objects.Sequences
                else:
                    cls = objects.View
                    coll_cls = objects.Views
                coll = meta.find_exact(cls=coll_cls,
                                       parent=schema)
                if coll is None:
                    coll = coll_cls(meta, parent=schema)
                    meta.set_object(coll)
                obj = cls(meta, schema=schema,
                          oid=item['reloid'], name=item['relname'],
                          comment=item['description'],
                          parent=coll)
                idxs = objects.Indexes(meta, parent=obj)
                meta.set_object(idxs)
                meta.set_object(obj)
            elif item['objtype'] == 'user':
                user = objects.User(meta, name=item['relname'],
                                    parent=users)
                meta.set_object(user)

    def refresh(self, obj, meta, connection):
        if obj.typeid == 'columns':
            self._refresh_columns(obj, meta, connection)
        elif obj.typeid == 'languages':
            self._refresh_languages(obj, meta, connection)
        elif obj.typeid == 'functions':
            self._refresh_functions(obj, meta, connection)
        elif obj.typeid == 'indexes':
            self._refresh_indexes(obj, meta, connection)
        elif obj.typeid == 'constraints':
            self._refresh_constraints(obj, meta, connection)
        # View definitions: select pg_get_viewdef(%(oid)s, true)

    def _refresh_columns(self, coll, meta, connection):
        table = coll.parent
        sql = ("select att.attnum, att.attname, dsc.description"
               " from pg_attribute att"
               " left join pg_description dsc on dsc.objoid = %(tableoid)s"
               "  and dsc.objsubid = att.attnum"
               " where att.attrelid = %(tableoid)s"
               " and att.attnum >= 1"
               % {"tableoid" : table.get_data("oid")})
        known_columns = {}
        [known_columns.setdefault(k.name, k)
         for k in meta.find(parent=coll, cls=objects.Column)]
        for item in self._query(connection, sql):
            col = known_columns.get(item['attname'], None)
            if col is None:
                col = objects.Column(meta, parent=coll)
            col.name = item['attname']
            col.pos = item['attnum']
            col.comment = item['description']
            meta.set_object(col)

    def _refresh_constraints(self, coll, meta, connection):
        sql = ("select con.oid, con.conname, dsc.description"
               " from pg_constraint con"
               " left join pg_description dsc on dsc.objoid = con.oid"
               " where con.conrelid = %d" % coll.parent.oid)
        for item in self._query(connection, sql):
            con = meta.find_exact(parent=coll, oid=item['oid'])
            if con is None:
                con = objects.Constraint(meta, parent=coll, oid=item['oid'])
                meta.set_object(con)
            con.name = item['conname']
            con.comment = item['description']

    def _refresh_indexes(self, coll, meta, connection):
        sql = ("select rel.oid, rel.relname, dsc.description"
               " from pg_index ind, pg_class rel"
               " left join pg_description dsc on dsc.objoid = rel.oid"
               " where ind.indexrelid = rel.oid"
               " and ind.indrelid = %d" % coll.parent.oid)
        for item in self._query(connection, sql):
            idx = meta.find_exact(parent=coll, oid=item['oid'])
            if idx is None:
                idx = objects.Index(meta, parent=coll, oid=item['oid'])
                meta.set_object(idx)
            idx.name = item['relname']
            idx.comment = item['description']

    def _refresh_languages(self, coll, meta, connection):
        sql = ("select lan.oid, lan.lanname, dsc.description"
               " from pg_language lan"
               " left join pg_description dsc"
               "  on dsc.objoid = lan.oid")
        for item in self._query(connection, sql):
            lan = meta.find_exact(parent=coll, oid=item['oid'])
            if lan is None:
                lan = objects.Language(meta, parent=coll,
                                       oid=item['oid'])
                meta.set_object(lan)
            lan.name = item['lanname']
            lan.comment = item['description']

    def _refresh_functions(self, coll, meta, connection):
        sql = ("select pro.oid, pro.proname, dsc.description"
               " from pg_proc pro"
               " left join pg_description dsc"
               "  on dsc.objoid = pro.oid"
               " where pro.pronamespace = %d" % coll.parent.oid)
        for item in self._query(connection, sql):
            pro = meta.find_exact(parent=coll, oid=item["oid"])
            if pro is None:
                pro = objects.Function(meta, parent=coll, oid=item["oid"])
                meta.set_object(pro)
            pro.name = item["proname"]
            pro.comment = item["description"]

    def get_transaction_state(self, connection):
        # The postgres backend retrieves this information from the
        # DB-API2 connection.
        psycopg2 = self.dbapi()
        conn = connection.get_dbapi_connection()
        stat = conn.get_transaction_status()
        if stat == psycopg2.extensions.TRANSACTION_STATUS_INERROR:
            flag = TRANSACTION_ROLLBACK_ENABLED
        elif stat == psycopg2.extensions.TRANSACTION_STATUS_ACTIVE:
            flag = TRANSACTION_IDLE
        elif stat == psycopg2.extensions.TRANSACTION_STATUS_IDLE:
            flag = TRANSACTION_IDLE
        elif stat == psycopg2.extensions.TRANSACTION_STATUS_INTRANS:
            flag = TRANSACTION_COMMIT_ENABLED|TRANSACTION_ROLLBACK_ENABLED
        else:
            flag = TRANSACTION_IDLE
        return flag

    def begin(self, connection):
        connection.execute('BEGIN')
        return True

DRIVER = Postgres


PG_INITIAL_SQL = """
SELECT nsp.oid AS nspoid,
       nsp.nspname,
       rel.oid AS reloid,
       rel.relname,
       des.description,
       CASE WHEN rel.relkind = 'r' THEN 'table'
            WHEN rel.relkind = 'i' THEN 'index'
            WHEN rel.relkind = 'S' THEN 'sequence'
            WHEN rel.relkind = 'v' THEN 'view'
            ELSE rel.relkind::char  -- somehow this is required, otherwise the above fails
       END AS objtype --        rel.*

FROM pg_class rel
LEFT JOIN pg_description des ON des.objoid = rel.oid and des.objsubid is null
LEFT JOIN pg_namespace nsp ON nsp.oid = rel.relnamespace
WHERE rel.relkind IN ('v', 'r', 'S')

UNION

-- languages
SELECT NULL, NULL, lan.oid,
                   lan.lanname,
                   des.description, 'language'
FROM pg_language lan
LEFT JOIN pg_description des ON des.objoid = lan.oid

UNION

-- users
SELECT NULL, NULL, use.usesysid,
                   use.usename, NULL, 'user'
FROM pg_user use

union

-- functions
select nsp.oid, nsp.nspname, pro.oid, pro.proname, des.description, 'function'
from pg_proc pro
left join pg_namespace nsp on nsp.oid = pro.pronamespace
left join pg_description des on des.objoid = pro.oid
"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Main application window"""

import logging
import os
import urlparse
import webbrowser
from gettext import gettext as _

import gobject
import gtk

from cf import USER_CONFIG_DIR
from cf import release
from cf.db import TRANSACTION_IDLE, TRANSACTION_COMMIT_ENABLED
from cf.db import TRANSACTION_ROLLBACK_ENABLED
from cf.ui import pane
from cf.ui import utils
from cf.ui import widgets
from cf.ui.browser import Browser
from cf.ui.confirmsave import ConfirmSaveDialog
from cf.ui.datasources import DatasourcesDialog
from cf.ui.editor import Editor
from cf.ui.statusbar import CrunchyStatusbar


class MainWindow(gtk.Window):
    """Application main window.

    Instance attributes:
      menubar: Menu bar.

    :Signals:
    active-editor-changed
      ``def callback(instance, editor)``
      Emitted when the active editor has changed. editor is either an
      editor instance or None.
    """

    __gsignals__ = {
        'active-editor-changed': (gobject.SIGNAL_RUN_LAST,
                                  gobject.TYPE_NONE,
                                  (gobject.TYPE_PYOBJECT,)),
        'editor-created': (gobject.SIGNAL_RUN_LAST,
                           gobject.TYPE_NONE,
                           (gobject.TYPE_PYOBJECT,)),
    }

    def __init__(self, app, create_editor=True):
        gtk.Window.__init__(self)
        self.app = app
        self._editor = None
        self._editor_conn_tag = None  # Signal ID
        self._tracked_conn = None  # Tracked connection
        self._editors = []
        self.clipboard = gtk.clipboard_get()
        self.ui = gtk.UIManager()
        # See: http://www.daa.com.au/pipermail/pygtk/2006-May/012267.html
        self.ui.connect('connect-proxy', self.on_ui_connect_proxy)
        self.ui.connect('disconnect-proxy', self.on_ui_disconnect_proxy)
        self._init_actions()
        self.ui.add_ui_from_string(UI)
        self.ui.ensure_update()
        self.set_title("CrunchyFrog %s" % release.version)
        self.set_icon_list(*utils.get_logo_icon_list())
        self._init_elements()
        self.connect('window-state-event', self.on_window_state_event)
        self.connect('delete-event', self.on_quit)
        # XXX disabled: it results in unexpected behavior...
        if create_editor and 1 == 2:
            self.editor_create()
        self.state_restore()

    def _init_actions(self):
        group = gtk.ActionGroup('instance')
        group.set_data('cf::label', _(u'Application'))
        entries = (
            # File
            ('file-menu-action', None, _(u'_File')),
            ('instance-quit', gtk.STOCK_QUIT,
             None, '<control>Q', _(u'Quit the program'),
             self.on_quit),
            ('file-new-menu-action', None, _(u'_New')),
            ('instance-new-editor', gtk.STOCK_NEW,
             _(u'_Query'), '<control>N', _(u'New SQL editor'),
               self.on_editor_new),
            ('instance-new-mainwin', None,
             _(u'_Window'), '<shift><control>N',
             _('New application window.'),
             self.on_instance_new),
            ('file-open', gtk.STOCK_OPEN,
             None, '<control>O', _(u'Open a file'),
             self.on_open_file),
            # Edit
            ('edit-menu-action', None, _(u'_Edit')),
            ('instance-plugins', None,
             _(u'_Plugins'), None, _(u'Configure plugins'),
             lambda action: self.app.preferences_show(self, 'plugins')),
            ('instance-datasourcemanager', None, _(u'_Data Sources'),
             None, _(u'Add, edit and delete data sources'),
             self.on_datasource_manager),
            ('instance-preferences', gtk.STOCK_PREFERENCES,
             None, None, _(u'Configure the application'),
             lambda action: self.app.preferences_show(self)),
            # Query
            ('query-menu-action', None, _(u'_Query')),
            # View
            ('view-menu-action', None, _(u'_View')),
            # Help
            ('help-menu-action', None, _(u'_Help')),
            ('help-help', gtk.STOCK_HELP,
             None, 'F1', _(u'Open the CrunchyFrog manual'),
             lambda *a: self.app.show_help()),
            ('help-translate', None,
             _(u'Help translate...'), None,
             _(u'Help to translate this application'),
             lambda a:
             webbrowser.open('https://translations.launchpad.net/crunchyfrog'),
             ),
            ('help-bugs', None,
             _(u'Report a problem'), None,
             _((u'Help to improve this application by reporting a bug '
                u'or a feature request.')),
             lambda a:
             webbrowser.open(('http://code.google.com/p/crunchyfrog/'
                              'issues/entry'))),
            ('help-about', gtk.STOCK_ABOUT,
             None, None, _(u'About this application'),
             self.on_about),
        )
        group.add_actions(entries)
        toggle_entries = (
            ('instance-toggle-toolbar', None,
             _(u'_Toolbar'), None,
             _(u'Show or hide the toolbar in the current window'),
             self.on_toggle_toolbar),
            ('instance-toggle-statusbar', None,
             _(u'_Statusbar'), None,
             _(u'Show or hide the statusbar in the current window'),
             self.on_toggle_statusbar),
        )
        group.add_toggle_actions(toggle_entries)
        # Editor switcher
        entries = []
        for i in range(1, 11):
            if i == 10:
                i = 0
            entry = ('activate-editor%d' % i, None, '', '<Alt>%d' % i, None,
                     i-1)
            entries.append(entry)
        group.add_radio_actions(entries, 0, self.on_editor_set_focus)
        self.ui.insert_action_group(group, -1)
        group = gtk.ActionGroup('editor')
        group.set_data('cf::label', _(u'Editor'))
        entries = (
            ('editor-close', gtk.STOCK_CLOSE,
             None, '<control>W', _(u'Close the current editor'),
             self.on_editor_close),
            ('editor-close-all', None,
             _(u'Close _all'), '<control><shift>W', _(u'Close all editors'),
             lambda action: self.close_all_editors()),
            # Query
            ('query-connection-menu-action', None, _(u'_Connection')),
            ('query-connection-disconnect', gtk.STOCK_DISCONNECT,
             None, '<shift>F8',
             _(u'Disconnect current editor'),
             self.on_editor_disconnect),
            ('query-show-connections', None,
             _(u'Show connections'), '<shift>F5',
             _(u'Open or close connections in a dialog'),
             lambda *a: self.app.manage_connections(self)),
            ('query-next-statement', None,
             _(u'Next Statement'), '<control><alt>j',
             _(u'Move cursor to next statement'),
             self.on_jump_next_statement),
            ('query-prev-statement', None,
             _(u'Previous Statement'), '<control><alt>k',
             _(u'Move cursor to previous statement'),
             self.on_jump_prev_statement),
            ('query-frmt-menu', None, _(u'_Format')),
            ('query-frmt-comment', None,
             _(u'_Comment / Uncomment'), '<shift><control>space',
             _(u'Comment / uncomment selected lines'),
             self.on_frmt_comment),
            ('query-frmt-format', None,
             _(u'_Format'), '<shift><control>f',
             _(u'Beautify selected lines or whole buffer'),
             self.on_frmt_format),
            ('query-frm-uppercase-kw', None,
             _(u'_Uppercase Keywords'), '<shift><control>u',
             _(u'Uppercase keywords in selected lines or in whole buffer'),
             self.on_frmt_uppercase_kw),
            ('query-toggle-results', None,
             _(u'Show / Hide Results'), '<shift>F9',
             _(u'Show or hide results pane'),
             self.on_editor_toggle_results),
            ('file-save', gtk.STOCK_SAVE,
             None, '<control>S', _(u'Save the current file'),
             self.on_file_save),
            ('file-save-as', gtk.STOCK_SAVE_AS,
             None, '<shift><control>S',
             _(u'Save the current file with a different name'),
             self.on_file_save_as),
            ('editor-print', gtk.STOCK_PRINT,
             None, '<control>P', _(u'Print the current page'),
             self.on_editor_print),
            ('editor-printpreview', gtk.STOCK_PRINT_PREVIEW,
             None, '<shift><control>P', _(u'Print preview'),
             self.on_editor_print_preview),
        )
        group.add_actions(entries)
        group.set_sensitive(False)
        self.ui.insert_action_group(group, 0)
        # Clipboard
        group = gtk.ActionGroup('clipboard')
        entries = (
            ('clipboard-cut', gtk.STOCK_CUT,
             None, '<control>X', _(u'Cut the selection'),
             self.on_clipboard_cut),
            ('clipboard-copy', gtk.STOCK_COPY,
             None, '<control>C', _(u'Copy the selection'),
             self.on_clipboard_copy),
            ('clipboard-paste', gtk.STOCK_PASTE,
             None, '<control>V', _(u'Paste the clipboard'),
             self.on_clipboard_paste),
            )
        group.add_actions(entries)
        self.ui.insert_action_group(group, 0)
        group = gtk.ActionGroup('query')
        group.set_data('cf::label', _(u'Queries'))
        entries = (
            ('query-execute', gtk.STOCK_EXECUTE,
             None, 'F5', _(u'Execute statements in SQL editor'),
             self.on_query_execute),
            ('query-execute-current', None,
             _(u'Exec_ute Current Statement'), '<control>F5',
             _(u'Executes statement at cursor'),
             self.on_query_execute_current),
            ('query-begin', gtk.STOCK_INDENT,
             _(u'Transaction'), 'F6',
             _(u'Begin transaction on current connection'),
             self.on_begin_transaction),
            ('query-commit', gtk.STOCK_APPLY,
             _(u'Commit'), 'F7',
             _(u'Commit current transaction'),
             self.on_commit),
            ('query-rollback', gtk.STOCK_UNDO,
             _(u'Rollback'), 'F8',
             _(u'Rollback current transaction'),
             self.on_rollback),
        )
        group.add_actions(entries)
        group.set_sensitive(False)
        self.ui.insert_action_group(group, -1)
        self.add_accel_group(self.ui.get_accel_group())

    def _init_elements(self):
        """Initialize main window elements."""
        self._init_statusbar()
        self._init_panes()
        vbox = gtk.VBox()
        self.add(vbox)
        vbox.pack_start(self._init_menubar(), False, False)
        vbox.pack_start(self._init_toolbar(), False, False)
        hpaned = gtk.HPaned()
        vbox.pack_start(hpaned, True, True)
        hpaned.show()
        vpaned = gtk.VPaned()
        hpaned.add2(vpaned)
        vpaned.show()
        hpaned.add1(self.side_pane)
        vpaned.add2(self.bottom_pane)
        self.queries = pane.CenterPane(self)
        vpaned.add1(self.queries)
        self.queries.show()
        # Connect to realize to set paned position when window is ready.
        self.connect('realize', self.on_set_paned_position, vpaned)
        self.connect('realize', self.on_set_paned_position, hpaned)
        vbox.pack_start(self.statusbar, False, False)
        vbox.show()
        self.browser = Browser(self.app, self)
        self.side_pane.add_item(self.browser)
        self.side_pane.set_active_item('navigator')
        menu = self.ui.get_widget('/MenuBar/Query/Connection')
        menu.connect('activate', self.on_connection_menu_activate)
        menu = self.ui.get_widget('/MenuBar/Query')
        menu.connect('activate', self.on_query_menu_activate)
        self._init_file_open()

    def _init_file_open(self):
        """Initialize file open menu item and toolbar button."""
        ph = self.ui.get_widget("/ToolBar/OpenFile")
        btn = gtk.MenuToolButton(gtk.STOCK_OPEN)
        btn.show()
        btn.connect('clicked', self.on_open_file)
        btn.set_menu(self._get_recent_menu())
        idx = self.toolbar.get_item_index(ph)
        self.toolbar.insert(btn, idx)
        ph = self.ui.get_widget('/MenuBar/File/RecentFiles')
        menu = self._get_recent_menu()
        menuitem = gtk.MenuItem(_(u'_Recent Files'))
        menuitem.show()
        menuitem.set_submenu(menu)
        file_menu = self.ui.get_widget('/MenuBar/File').get_submenu()
        for idx, child in enumerate(file_menu.get_children()):
            if child == ph:
                file_menu.insert(menuitem, idx)
                return

    def _init_menubar(self):
        """Create and return the applications menubar."""
        self.menubar = self.ui.get_widget('/MenuBar')
        self.menubar.show_all()
        return self.menubar

    def _init_panes(self):
        """Init side and bottom pane."""
        self.side_pane = pane.SidePane(self)
        self.bottom_pane = pane.BottomPane(self)

    def _init_statusbar(self):
        """Create and return the statusbar."""
        self.statusbar = CrunchyStatusbar(self.app, self)
        self.statusbar.show()
        return self.statusbar

    def _init_toolbar(self):
        """Create and return the toolbar."""
        self.toolbar = self.ui.get_widget("/ToolBar")
        self.toolbar.show_all()
        ph = self.ui.get_widget("/ToolBar/EditorConnection")
        self.tb_conn_chooser = widgets.ConnectionButton(self)
        idx = self.toolbar.get_item_index(ph)
        self.toolbar.insert(self.tb_conn_chooser, idx)
        self.tb_conn_chooser.show_all()
        separator = gtk.SeparatorToolItem()
        self.toolbar.insert(separator, idx+1)
        separator.show()
        return self.toolbar

    def _get_action(self, action_name):
        """Return a action regardless of groups."""
        for group in self.ui.get_action_groups():
            action = group.get_action(action_name)
            if action is not None:
                return action
        return None

    def _get_clipboard(self):
        return self.clipboard

    def _get_recent_menu(self, limit=None, recent_menu=None):
        """Return a recent file menu."""
        if recent_menu is None:
            recent_menu = gtk.RecentChooserMenu(self.app.recent_manager)
        filter_ = gtk.RecentFilter()
        filter_.add_mime_type("text/x-sql")
        recent_menu.add_filter(filter_)
        recent_menu.set_filter(filter_)
        recent_menu.set_show_not_found(False)
        recent_menu.set_sort_type(gtk.RECENT_SORT_MRU)
        recent_menu.connect("item-activated", self.on_recent_item_activated)
        if limit is not None:
            recent_menu.set_limit(limit)
        return recent_menu

    # ---
    # Callbacks
    # ---
    def on_about(self, *args):
        def open_url(dialog, url):
            gtk.show_uri(dialog.get_screen(), url,
                         gtk.gdk.x11_get_server_time(dialog.window))
        gtk.about_dialog_set_url_hook(open_url)
        dlg = gtk.AboutDialog()
        dlg.set_name(release.name)
        dlg.set_version(release.version)
        dlg.set_copyright(release.copyright)
        dlg.set_license(release.license)
        dlg.set_website(release.url)
        dlg.set_website_label(release.url)
        dlg.set_logo(utils.get_logo_icon(96))
        dlg.set_program_name(release.appname)
        dlg.set_translator_credits(release.translators)
        dlg.run()
        dlg.destroy()


    def on_begin_transaction(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.begin_transaction)

    def on_clipboard_copy(self, *args):
        if not self._editor:
            return
        self._editor.clipboard_copy(self._get_clipboard())

    def on_clipboard_cut(self, *args):
        if not self._editor:
            return
        self._editor.clipboard_cut(self._get_clipboard())

    def on_clipboard_paste(self, *args):
        if not self._editor:
            return
        self._editor.clipboard_paste(self._get_clipboard())

    def on_commit(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.commit)

    def on_connection_menu_activate(self, menuitem):
        widgets.rebuild_connection_menu(menuitem.get_submenu(), self,
                                        self.get_active_editor())

    def on_connection_notify(self, connection, property):
        if property.name == 'transaction-state':
            value = connection.get_property(property.name)
            gobject.idle_add(self.set_transaction_state, value, connection)

    def on_datasource_manager(self, *args):
        dlg = DatasourcesDialog(self.app, self)
        dlg.run()
        dlg.destroy()

    def on_editor_close(self, *args):
        editor = self.get_active_editor()
        if editor is not None:
            editor.close()

    def on_editor_connection_changed(self, editor, connection):
        if self._tracked_conn:
            sig = self._tracked_conn.get_data('cf::mainwin::notify')
            if sig:
                self._tracked_conn.disconnect(sig)
            self._tracked_conn = None
        for group in self.ui.get_action_groups():
            if group.get_name() == 'query':
                group.set_sensitive(connection is not None)
        if connection:
            self.set_title(connection.get_label()+' - CrunchyFrog')
        else:
            self.set_title('CrunchyFrog %s' % release.version)
        # Set transaction state
        if connection:
            self._tracked_conn = connection
            sig = self._tracked_conn.connect('notify',
                                             self.on_connection_notify)
            self._tracked_conn.set_data('cf::mainwin::notify', sig)
            state = connection.get_property('transaction-state')
            self.set_transaction_state(state, connection)

    def on_editor_disconnect(self, action):
        editor = self.get_active_editor()
        if editor is None or editor.connection is None:
            return
        editor.set_connection(None)

    def on_editor_new(self, *args):
        editor = self.editor_create()
        # FIXME(andi): Updating editor switcher action should be in a
        #              separate function.
        self.on_query_menu_activate(None)

    def on_editor_print(self, action):
        self.get_active_editor().print_contents()

    def on_editor_print_preview(self, action):
        self.get_active_editor().print_contents(preview=True)

    def on_editor_set_focus(self, first_action, current):
        queries_idx = current.get_current_value()
        if queries_idx == -1:
            queries_idx = 9
        if self.queries.get_n_pages()-1 < queries_idx:
            return
        self.queries.set_current_page(queries_idx)
        editor = self.queries.get_nth_page(queries_idx)
        focus_child = editor.get_focus_child()
        if focus_child is not None:
            focus_child.grab_focus()

    def on_editor_toggle_results(self, action):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.toggle_results_pane()

    def on_frmt_comment(self, action):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.selected_lines_toggle_comment()

    def on_frmt_format(self, action):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.selected_lines_quick_format()

    def on_frmt_uppercase_kw(self, action):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.selected_lines_quick_format(keyword_case='upper')

    def on_file_save(self, *args):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.save_file(self)

    def on_file_save_as(self, *args):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.save_file_as(self)

    def on_instance_new(self, action):
        self.app.new_instance()

    def on_jump_next_statement(self, action):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.rjump_to_statement(1)

    def on_jump_prev_statement(self, action):
        editor = self.get_active_editor()
        if editor is None:
            return
        editor.rjump_to_statement(-1)

    def on_menu_item_deselect(self, menuitem):
        self.statusbar.pop(100)

    def on_menu_item_select(self, menuitem, tooltip):
        self.statusbar.push(100, tooltip)

    def on_open_file(self, *args):
        dlg = gtk.FileChooserDialog(_(u"Select file"),
                            self,
                            gtk.FILE_CHOOSER_ACTION_OPEN,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
        dlg.set_select_multiple(True)
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        recent_chooser = gtk.RecentChooserWidget(self.app.recent_manager)
        filter = gtk.RecentFilter()
        filter.add_mime_type("text/x-sql")
        filter.set_name(_(u"SQL files (*.sql)"))
        recent_chooser.add_filter(filter)
        recent_chooser.set_filter(filter)
        recent_chooser.set_show_not_found(False)
        recent_chooser.set_sort_type(gtk.RECENT_SORT_MRU)
        recent_chooser.set_show_icons(True)
        recent_chooser.set_show_tips(True)
        def dlg_set_uri(chooser, dlg):
            if chooser.get_current_uri():
                dlg.set_uri(chooser.get_current_uri())
        recent_chooser.connect("selection-changed", dlg_set_uri, dlg)
        recent_chooser.connect("item-activated",
                               lambda chooser: dlg.response(gtk.RESPONSE_OK))
        exp = gtk.Expander(_(u"_Recent files:"))
        exp.add(recent_chooser)
        exp.set_use_underline(True)
        dlg.set_extra_widget(exp)
        recent_chooser.show()
        if dlg.run() == gtk.RESPONSE_OK:
            [self.editor_create(fname) for fname in dlg.get_filenames()]
            self.app.config.set("editor.recent_folder",
                                dlg.get_current_folder())
        dlg.destroy()

    def on_preferences(self, *args):
        self.app.preferences_show()

    def on_query_execute(self, *args):
        self.get_active_editor().execute_query()

    def on_query_execute_current(self, action):
        self.get_active_editor().execute_query(True)

    def on_query_menu_activate(self, menuitem):
        self._rebuild_activate_editor_actions()

    def _rebuild_activate_editor_actions(self):
        for idx in range(1, 11):
            if idx == 10:
                idx = 0
                page_idx = 9
            else:
                page_idx = idx - 1
            action = self._get_action('activate-editor%d' % idx)
            page = self.queries.get_nth_page(page_idx)
            if page is None:
                action.set_visible(False)
                action.set_sensitive(False)
            else:
                action.set_visible(True)
                action.set_sensitive(True)
                tab = self.queries.get_tab_label(page)
                lbl = tab.label.get_text()
                if len(lbl) > 16:
                    lbl = lbl[:15]+'...'
                action.props.label = lbl.replace('_', '__')
                is_active = bool(page_idx == self.queries.get_current_page())
                action.set_active(is_active)

    def on_quit(self, *args):
        if not self.close_all_editors():
            return False
        if len(self.app.get_instances()) <= 1:
            self.state_save()
        self.destroy()

    def on_recent_item_activated(self, chooser):
        self.editor_create(chooser.get_current_uri())

    def on_rollback(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.rollback)

    def on_set_paned_position(self, win, paned):
        if isinstance(paned, gtk.VPaned):
            opt = 'win.bottompane_position'
        else:
            opt = 'win.sidepane_position'
        w, h = win.get_size()
        stored_pos = self.app.config.get(opt, None)
        if stored_pos is None:
            if isinstance(paned, gtk.VPaned):
                stored_pos = h-250
            else:
                stored_pos = 250
        paned.set_position(stored_pos)
        paned.connect('notify::position',
                      lambda paned, prop:
                      self.app.config.set(opt, paned.get_position()))

    def on_toggle_statusbar(self, toggle_action):
        if toggle_action.get_active():
            self.statusbar.show()
        else:
            self.statusbar.hide()
        self.app.config.set('win.statusbar_visible',
                            toggle_action.get_active())

    def on_toggle_toolbar(self, toggle_action):
        if toggle_action.get_active():
            self.toolbar.show()
        else:
            self.toolbar.hide()
        self.app.config.set('win.toolbar_visible',
                            toggle_action.get_active())

    def on_ui_connect_proxy(self, ui, action, widget):
        tooltip = action.get_property('tooltip')
        if isinstance(widget, gtk.MenuItem) and tooltip:
            cid = widget.connect('select', self.on_menu_item_select, tooltip)
            cid2 = widget.connect('deselect', self.on_menu_item_deselect)
            widget.set_data('cf::cids', (cid, cid2))

    def on_ui_disconnect_proxy(self, ui, action, widget):
        cids = widget.get_data('cf::cids') or ()
        for cid in cids:
            widget.disconnect(cid)

    def on_window_state_event(self, window, event):
        val_names = event.new_window_state.value_names
        c = self.app.config
        c.set('gui.maximized',
              gtk.gdk.WINDOW_STATE_MAXIMIZED.value_names[0] in val_names)

    # ---
    # Public methods
    # ---

    def close_all_editors(self):
        """Closes all editors but checks for changes first.

        Returns:
          False if action was cancelled, otherwise True.
        """
        changed = []
        for item in self.queries.get_all_editors():
            if isinstance(item, Editor) and item.contents_changed():
                changed.append(item)
        if changed:
            dlg = ConfirmSaveDialog(self, changed)
            proceed = dlg.run()
            if proceed == 1:
                dlg.hide()
                if not dlg.save_files():
                    proceed = 0
            dlg.destroy()
            if proceed == 0:
                return False
        for item in self.queries.get_all_editors():
            item.close(force=True)
        return True

    def get_editors(self):
        """Return a list of all SQL editors."""
        return [item for item in self.queries.get_all_editors()
                if isinstance(item, Editor)]

    def editor_create(self, fname=None):
        """Creates a new SQL editor.

        Arguments:
          fname: If given, the file fname is opened with this editor.

        Returns:
          An Editor instance.
        """
        last_conn = None
        if self.app.config.get('editor.reuse_connection'):
            cur_editor = self.get_active_editor()
            if cur_editor is not None:
                last_conn = cur_editor.get_connection()
        editor = Editor(self)
        if last_conn is not None:
            editor.set_connection(last_conn)
        if fname:
            scheme, _, path, _, _, _ = urlparse.urlparse(fname)
            fname = path
            editor.set_filename(fname)
        self.editor_append(editor)
        editor.show_all()
        editor.textview.grab_focus()
        return editor

    def editor_append(self, editor):
        """Adds an editor to the editors notebook.

        Arguments:
          editor: Editor instance.
        """
        # Cleanup other instances
        for instance in self.app.get_instances():
            if instance != self and editor in instance._editors:
                instance._editors.remove(editor)
        self.queries.add_item(editor)
        self._editors.append(editor)

    def editor_remove(self, editor):
        """Removes an editor from this instance."""
        if editor in self._editors:
            self._editors.remove(editor)

    def get_active_editor(self):
        """Returns the active editor.

        Returns:
          Editor instance or None.
        """
        if isinstance(self._editor, Editor):
            return self._editor
        return None

    def set_editor_active(self, editor, active):
        """Called whenever an editor receives or looses focus."""
        if not active:
            editor = None
        if self._editor_conn_tag and self._editor:
            self._editor.disconnect(self._editor_conn_tag)
            self._editor_conn_tag = None
        if self._editor:
            handler_id = self._editor.get_data('cf::sig_editor_buffer_changed')
            if handler_id:
                self._editor.disconnect(handler_id)
                self._editor.set_data('cf::sig_editor_buffer_changed', None)
        self._editor = editor
        if self._editor and isinstance(self._editor, Editor):
            self._editor_conn_tag = self._editor.connect(
                "connection-changed", self.on_editor_connection_changed)
            self.on_editor_connection_changed(
                self._editor, self._editor.connection)
            conn = self._editor.connection
            if conn:
                prop =conn.get_property('transaction-state')
                self.set_transaction_state(prop, conn)
            # XXX add buffer selection changed cb to update copy&paste btns
        else:
            self.set_title('CrunchyFrog %s' % release.version)
        for group in self.ui.get_action_groups():
            if group.get_name() == 'editor':
                group.set_sensitive(isinstance(self._editor, Editor))
            elif group.get_name() == 'query':
                group.set_sensitive(bool(isinstance(self._editor, Editor)
                                         and self._editor.connection))
        sensitive = bool((self._editor
                          and isinstance(self._editor, Editor)))
        action = self._get_action('file-save')
        action.set_sensitive(sensitive)
        action = self._get_action('file-save-as')
        action.set_sensitive(sensitive)
        self.tb_conn_chooser.set_editor(editor)
        self.app.plugins.editor_notify(editor, self)
        self.emit('active-editor-changed', editor)
        if editor is not None and isinstance(editor, Editor):
            editor.textview.grab_focus()

    def set_transaction_state(self, value, connection):
        """Adjusts the transactions state in the UI."""
        # A regression: If value is None that means we have no connection,
        # but this is already handled by setting the whole action group
        # insensitive.
        if value is None or connection is None:
            return
        commit = self._get_action('query-commit')
        rollback = self._get_action('query-rollback')
        begin = self._get_action('query-begin')
        commit.set_sensitive((value & TRANSACTION_COMMIT_ENABLED) != 0)
        rollback.set_sensitive((value & TRANSACTION_ROLLBACK_ENABLED) != 0)
        begin.set_sensitive((value & TRANSACTION_IDLE) != 0)

    def state_restore(self):
        """Restore window state."""
        conf = self.app.config
        action = self._get_action('instance-toggle-statusbar')
        action.set_active(conf.get('win.statusbar_visible', True))
        self.on_toggle_statusbar(action)
        action = self._get_action('instance-toggle-toolbar')
        action.set_active(conf.get('win.toolbar_visible', True))
        self.on_toggle_toolbar(action)
        self.side_pane.state_restore()
        self.bottom_pane.state_restore()
        width = conf.get('gui.width')
        height = conf.get('gui.height')
        if not width or width == -1:
            width = 800
        if not height or height == -1:
            height = 600
        self.resize(width, height)
        if conf.get('gui.maximized', False):
            self.maximize()
        fname = os.path.join(USER_CONFIG_DIR, 'shortcuts.map')
        if os.path.isfile(fname):
            gtk.accel_map_load(fname)

    def state_save(self):
        """Save window state to config."""
        c = self.app.config
        w, h = self.get_size()
        c.set('gui.width', w)
        c.set('gui.height', h)
        for pane in [self.side_pane, self.bottom_pane]:
            c.set('win.%s_visible' % pane.__class__.__name__.lower(),
                  pane.get_property('visible'))
        fname = os.path.join(USER_CONFIG_DIR, 'shortcuts.map')
        gtk.accel_map_save(fname)

gobject.type_register(MainWindow)


# NOTE: All menu actions MUST end with *-menu-action. That's needed by
#       the preferences dialog to filter out some actions.
#       Maybe there's a better way to do it, but it works... ;-)

UI = '''
  <menubar name="MenuBar">
    <menu name="File" action="file-menu-action">
      <menu name="FileNew" action="file-new-menu-action">
        <menuitem name="NewEditor" action="instance-new-editor" />
        <menuitem name="NewInstance" action="instance-new-mainwin" />
      </menu>
      <menuitem name="Open" action="file-open" />
      <menuitem name="Save" action="file-save" />
      <menuitem name="SaveAs" action="file-save-as" />
      <placeholder name="RecentFiles" />
      <separator />
      <menuitem name="Print" action="editor-print" />
      <menuitem name="PrintPreview" action="editor-printpreview" />
      <separator />
      <menuitem name="Close" action="editor-close" />
      <menuitem name="CloseAll" action="editor-close-all" />
      <menuitem name="Quit" action="instance-quit" />
    </menu>
    <menu name="Edit" action="edit-menu-action">
      <menuitem name="Cut" action="clipboard-cut" />
      <menuitem name="Copy" action="clipboard-copy" />
      <menuitem name="Paste" action="clipboard-paste" />
      <separator />
      <menuitem name="Plugins" action="instance-plugins" />
      <menuitem name="Datasources" action="instance-datasourcemanager" />
      <menuitem name="Preferences" action="instance-preferences" />
    </menu>
    <menu name="View" action="view-menu-action">
      <menuitem name="Toolbar" action="instance-toggle-toolbar" />
      <menuitem name="Statusbar" action="instance-toggle-statusbar" />
    </menu>
    <menu name="Query" action="query-menu-action">
      <menu name="Connection" action="query-connection-menu-action">
        <separator />
        <menuitem name="Disconnect" action="query-connection-disconnect" />
        <menuitem name="Connections" action="query-show-connections" />
      </menu>
      <separator />
      <menuitem name="Execute" action="query-execute" />
      <menuitem name="ExecuteCurrent" action="query-execute-current" />
      <menuitem name="Begin" action="query-begin" />
      <menuitem name="Commit" action="query-commit" />
      <menuitem name="Rollback" action="query-rollback" />
      <separator />
      <menuitem name="QueryJumpNext" action="query-next-statement" />
      <menuitem name="QueryJumpPrev" action="query-prev-statement" />
      <separator />
      <menu name="QueryFormat" action="query-frmt-menu">
        <menuitem name="QueryFormatComment" action="query-frmt-comment" />
        <menuitem name="QueryFormatFormat" action="query-frmt-format" />
        <menuitem name="QueryFormatKwUpper" action="query-frm-uppercase-kw" />
      </menu>
      <menuitem name="QueryToggleResults" action="query-toggle-results" />
      <placeholder name="query-extensions" />
      <separator />
      <menuitem name="Editor1" action="activate-editor1" />
      <menuitem name="Editor2" action="activate-editor2" />
      <menuitem name="Editor3" action="activate-editor3" />
      <menuitem name="Editor4" action="activate-editor4" />
      <menuitem name="Editor5" action="activate-editor5" />
      <menuitem name="Editor6" action="activate-editor6" />
      <menuitem name="Editor7" action="activate-editor7" />
      <menuitem name="Editor8" action="activate-editor8" />
      <menuitem name="Editor9" action="activate-editor9" />
      <menuitem name="Editor0" action="activate-editor0" />
    </menu>
    <menu name="Help" action="help-menu-action">
      <menuitem name="HelpHelp" action="help-help" />
      <separator />
      <menuitem name="HelpTranslate" action="help-translate" />
      <menuitem name="HelpBugs" action="help-bugs" />
      <menuitem name="HelpAbout" action="help-about" />
    </menu>
  </menubar>
  <toolbar name="ToolBar">
    <toolitem name="NewEditor" action="instance-new-editor" />
    <placeholder name="OpenFile" action="instance-open-file" />
    <toolitem name="SaveFile" action="file-save" />
    <separator />
    <placeholder name="EditorConnection" />
    <separator />
    <toolitem name="QueryExecute" action="query-execute" />
    <toolitem name="QueryBegin" action="query-begin" />
    <toolitem name="QueryCommit" action="query-commit" />
    <toolitem name="QueryRollback" action="query-rollback" />
  </toolbar>
'''



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""CrunchyFrog Package"""

import gettext
import os
import os.path
import sys
import urlparse


class CFError(Exception):
    """Base class for all errors."""


try:
    from local_config import PREFIX
except ImportError:
    PREFIX = '/usr/'


def _is_source_dir():
    """Checks if the application is started from source checkout/dist."""
    anchor = os.path.dirname(__file__)
    setup_py = os.path.join(anchor, '../setup.py')
    # Look into setup.py. It's not enough to check for the presence of
    # setup.py. Some modules (e.g. virtualenv) install a setup.py directly
    # in site-packages :-(
    if os.path.exists(setup_py) and 'crunchyfrog' in open(setup_py).read():
        return True
    return False

if _is_source_dir():
    DATA_DIR = os.path.join(os.path.dirname(__file__), '../data/')
    LOCALE_DIR = os.path.join(os.path.dirname(__file__), '../po/')
    MANUAL_URL = os.path.abspath(
        os.path.join(os.path.dirname(__file__),
                     '../docs/manual/build/html/')
        )
else:
    DATA_DIR = os.path.join(PREFIX, 'share/crunchyfrog/')
    LOCALE_DIR = os.path.join(PREFIX, 'share/locale/')
    MANUAL_URL = '/usr/share/doc/crunchyfrog/manual/'

if not sys.platform.startswith('win'):
    MANUAL_URL = urlparse.urlunsplit(('file', '', MANUAL_URL, '', ''))

DATA_DIR = os.path.abspath(DATA_DIR)
LOCALE_DIR = os.path.abspath(LOCALE_DIR)


PLUGIN_DIR = os.path.join(DATA_DIR, "plugins")

if not sys.platform.startswith('win'):
    from xdg import BaseDirectory as base_dir
    USER_CONFIG_DIR = base_dir.save_config_path('crunchyfrog')
    USER_DIR = base_dir.save_data_path('crunchyfrog')
else:
    USER_CONFIG_DIR = os.path.abspath(
        os.path.expanduser('~/crunchyfrog/config/'))
    USER_DIR = os.path.abspath(
        os.path.expanduser('~/crunchyfrog/data'))

USER_CONF = os.path.join(USER_CONFIG_DIR, "config")

if not os.path.isdir(USER_CONFIG_DIR):
    os.makedirs(USER_CONFIG_DIR)

if not os.path.isdir(USER_DIR):
    os.makedirs(USER_DIR)

USER_PLUGIN_DIR = os.path.join(USER_DIR, "plugins/")
if not os.path.isdir(USER_PLUGIN_DIR):
    os.makedirs(USER_PLUGIN_DIR)

IPC_SOCKET = os.path.join(USER_DIR, "crunchyfog.sock")

gettext.bindtextdomain('crunchyfrog', LOCALE_DIR)
gettext.textdomain('crunchyfrog')
gettext.install('crunchyfrog', LOCALE_DIR, True)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Firebird backend."""

from cf.db.backends import Generic, DEFAULT_OPTIONS


class Firebird(Generic):

    drivername = 'firebird'

    @classmethod
    def dbapi(cls):
        import kinterbasdb
        return kinterbasdb

    @classmethod
    def get_native_shell_command(cls, url):
        args = []
        if url.username:
            args.extend(['-u', url.username])
        if url.password:
            args.extend(['-p', url.password])
        args.extend([url.database])
        print args
        return 'isql-fb', args

    @classmethod
    def get_connect_params(cls, url):
        opts = url.get_dict()
        if 'username' in opts:
            opts['user'] = opts['username']
            del opts['username']
        if opts.get('port'):
            opts['host'] = "%s/%s" % (opts['host'], opts['port'])
            del opts['port']
        opts.update(url.query)

        if not cls.dbapi().initialized:
            cls.dbapi().init()

        return ([], opts)

    @classmethod
    def get_options(cls):
        return DEFAULT_OPTIONS

    def get_server_info(self, connection):
        return 'Firebird %s' % connection.connection.server_version

    def get_explain_statement(self, statement):
        return None


DRIVER = Firebird



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Custom statusbar"""

from gettext import gettext as _

import gtk

from cf.ui.editor import Editor


class CrunchyStatusbar(gtk.Statusbar):
    """An extended gtk.Statusbar."""

    def __init__(self, app, instance):
        """Constructor."""
        gtk.Statusbar.__init__(self)
        self.app = app
        self.instance = instance
        self._editor_sigs = {'connection-changed': None,
                             'toggle-overwrite': None,
                             'insert-mark': None}
        self._editor = None

        self.lbl_conn = gtk.Label()
        self.lbl_conn.set_alignment(0, 0.5)
        self.lbl_conn.show()
        self._add_item(self.lbl_conn, False, True)
        self.lbl_curpos = gtk.Label()
        self.lbl_curpos.set_alignment(1, 0.5)
        self.lbl_curpos.set_width_chars(15)
        self.lbl_curpos.show()
        self._add_item(self.lbl_curpos, False, True)
        self.lbl_insmode = gtk.Label()
        self.lbl_insmode.set_width_chars(4)
        self.lbl_insmode.show()
        self._add_item(self.lbl_insmode, False, True)

        self.instance.connect('active-editor-changed',
                              lambda i, e: self.set_editor(e))

    def _add_item(self, widget, expand=True, fill=True, padding=0):
        """Adds a widget to the statusbar using pack_start.

        Returns:
          The gtk.Frame holding the widget.
        """
        frame = gtk.Frame()
        frame.set_border_width(0)
        shadow_type = self.style_get_property('shadow_type')
        frame.set_property('shadow_type', shadow_type)
        frame.add(widget)
        frame.show()
        self.pack_start(frame, expand, fill, padding)
        return frame

    def _disconnect_editor_sigs(self):
        """Disconnects editor signals, if any."""
        if self._editor is not None:
            sigs = self._editor_sigs
            if sigs['connection-changed'] is not None:
                self._editor.disconnect(sigs['connection-changed'])
            if sigs['toggle-overwrite'] is not None:
                self._editor.textview.disconnect(sigs['toggle-overwrite'])
            if sigs['insert-mark'] is not None:
                self._editor.textview.buffer.disconnect(sigs['insert-mark'])
        self._editor_sigs['connection-changed'] = None
        self._editor_sigs['toggle-overwrite'] = None
        self._editor_sigs['insert-mark'] = None

    def _connect_editor_sigs(self):
        """Connect to some editor signals to track changes."""
        if self._editor is None or not isinstance(self._editor, Editor):
            return
        sig = self._editor.connect(
            'connection-changed',
            lambda e, c: self._set_connection_label(e))
        self._editor_sigs['connections-changed'] = sig
        sig = self._editor.textview.connect(
            'toggle-overwrite',
            lambda tv: self._set_overwrite_mode(self._editor, False))
        self._editor_sigs['toggle.overwrite'] = sig
        sig = self._editor.textview.buffer.connect('mark-set',
                                                   self.on_tv_mark_set)

    def _set_connection_label(self, editor):
        if editor is None:
            lbl = ''
        else:
            if editor.connection is not None:
                lbl = editor.connection.get_label()
            else:
                lbl = _(u'[Not connected]')
        self.lbl_conn.set_text(lbl)

    def _set_curpos(self, line=None, offset=None):
        if line is None or offset is None:
            lbl = ''
        else:
            lbl = _(u'Ln %(line)d, Col %(column)d') % {'line': line,
                                                       'column': offset}
        self.lbl_curpos.set_text(lbl)

    def _set_overwrite_mode(self, editor, toggle_bool=True):
        if editor is None:
            lbl = ''
        elif isinstance(editor, Editor):
            overwrite = editor.textview.get_overwrite()
            if toggle_bool:
                overwrite = not overwrite
            lbl = overwrite and _(u'INS') or _(u'OVR')
        else:
            lbl = ''
        self.lbl_insmode.set_text(lbl)

    # ---
    # Callbacks
    # ---

    def on_tv_mark_set(self, textbuffer, iter_, mark):
        mark = textbuffer.get_insert()
        iter_ = textbuffer.get_iter_at_mark(mark)
        self._set_curpos(iter_.get_line()+1, iter_.get_line_offset()+1)

    # ---
    # Public methods
    # ---

    def set_editor(self, editor):
        """Associates an editor with the statusbar.

        This method is (loosely) connected to the instance's
        'active-editor-changed' signal.

        Arguments:
          editor: Editor instance or None.
        """
        self._disconnect_editor_sigs()
        self.instance.statusbar.pop(1)
        self._editor = editor
        self._set_connection_label(self._editor)
        self._set_overwrite_mode(self._editor)
        if self._editor is None:
            self._set_curpos(None, None)
        elif isinstance(self._editor, Editor):
            self.on_tv_mark_set(self._editor.textview.buffer, None, None)
        if self._editor is None:
            return
        self._connect_editor_sigs()

    def set_message(self, msg, context=0):
        """Show an informal message in the statusbar.

        This method is just a wrapper for gtk.Statusbar.push(0, msg).

        Args:
          msg: Informal message (str).
        """
        self.push(context, msg)



# -*- coding: utf-8 -*-

"""build_manpage command -- Generate man page from setup()"""

import datetime
from distutils.command.build import build
from distutils.core import Command
from distutils.errors import DistutilsOptionError
import optparse


class build_manpage(Command):

    description = 'Generate man page from setup().'

    user_options = [
        ('output=', 'O', 'output file'),
        ('parser=', None, 'module path to optparser (e.g. mymod:func'),
        ]

    def initialize_options(self):
        self.output = None
        self.parser = None

    def finalize_options(self):
        if self.output is None:
            raise DistutilsOptionError('\'output\' option is required')
        if self.parser is None:
            raise DistutilsOptionError('\'parser\' option is required')
        mod_name, func_name = self.parser.split(':')
        fromlist = mod_name.split('.')
        try:
            mod = __import__(mod_name, fromlist=fromlist)
            self._parser = getattr(mod, func_name)()
        except ImportError, err:
            raise
        self._parser.formatter = ManPageFormatter()
        self._parser.formatter.set_parser(self._parser)
        self.announce('Writing man page %s' % self.output)
        self._today = datetime.date.today()

    def _markup(self, txt):
        return txt.replace('-', '\\-')

    def _write_header(self):
        appname = self.distribution.get_name()
        ret = []
        ret.append('.TH %s 1 %s\n' % (self._markup(appname),
                                      self._today.strftime('%Y\\-%m\\-%d')))
        description = self.distribution.get_description()
        if description:
            name = self._markup('%s - %s' % (self._markup(appname),
                                             description.splitlines()[0]))
        else:
            name = self._markup(appname)
        ret.append('.SH NAME\n%s\n' % name)
        synopsis = self._parser.get_usage()
        if synopsis:
            synopsis = synopsis.replace('%s ' % appname, '')
            ret.append('.SH SYNOPSIS\n.B %s\n%s\n' % (self._markup(appname),
                                                      synopsis))
        long_desc = self.distribution.get_long_description()
        if long_desc:
            ret.append('.SH DESCRIPTION\n%s\n' % self._markup(long_desc))
        return ''.join(ret)

    def _write_options(self):
        ret = ['.SH OPTIONS\n']
        ret.append(self._parser.format_option_help())
        return ''.join(ret)

    def _write_footer(self):
        ret = []
        appname = self.distribution.get_name()
        author = '%s <%s>' % (self.distribution.get_author(),
                              self.distribution.get_author_email())
        ret.append(('.SH AUTHORS\n.B %s\nwas written by %s.\n'
                    % (self._markup(appname), self._markup(author))))
        homepage = self.distribution.get_url()
        ret.append(('.SH DISTRIBUTION\nThe latest version of %s may '
                    'be downloaded from\n'
                    '.UR %s\n.UE\n'
                    % (self._markup(appname), self._markup(homepage),)))
        return ''.join(ret)

    def run(self):
        manpage = []
        manpage.append(self._write_header())
        manpage.append(self._write_options())
        manpage.append(self._write_footer())
        stream = open(self.output, 'w')
        stream.write(''.join(manpage))
        stream.close()


class ManPageFormatter(optparse.HelpFormatter):

    def __init__(self,
                 indent_increment=2,
                 max_help_position=24,
                 width=None,
                 short_first=1):
        optparse.HelpFormatter.__init__(self, indent_increment,
                                        max_help_position, width, short_first)

    def _markup(self, txt):
        return txt.replace('-', '\\-')

    def format_usage(self, usage):
        return self._markup(usage)

    def format_heading(self, heading):
        if self.level == 0:
            return ''
        return '.TP\n%s\n' % self._markup(heading.upper())

    def format_option(self, option):
        result = []
        opts = self.option_strings[option]
        result.append('.TP\n.B %s\n' % self._markup(opts))
        if option.help:
            help_text = '%s\n' % self._markup(self.expand_default(option))
            result.append(help_text)
        return ''.join(result)


build.sub_commands.append(('build_manpage', None))




# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Confirmation dialog."""

import gtk
import gobject

import os

from gettext import gettext as _


MODE_SINGLE = 1
MODE_MULTIPLE = 2


class ConfirmSaveDialog(object):

    def __init__(self, instance, changed_editors):
        """Constructor.

        :param instance: :class:`~cf.ui.MainWindow` instance.
        :param changed_editors: List of :class:`~cf.ui.editor.Editor` instances
          that have changed content.
        """
        app = instance.app
        self.instance = instance
        self.builder = gtk.Builder()
        self.builder.set_translation_domain('crunchyfrog')
        self.builder.add_from_file(app.get_glade_file('saveconfirm.glade'))
        self.builder.connect_signals(self)
        self.changed_editors = changed_editors
        self.dlg = self.builder.get_object('saveconfirm_dialog')
        self._setup_widget()
        if len(changed_editors) == 1:
            self.set_dialog_mode(MODE_SINGLE)
        else:
            self.set_dialog_mode(MODE_MULTIPLE)
        self.fill_filelist()

    def _setup_widget(self):
        model = gtk.ListStore(gobject.TYPE_PYOBJECT, str, bool, str)
        self.filelist = self.builder.get_object("filelist")
        self.filelist.set_model(model)
        self.filelist.set_tooltip_column(3)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererToggle()
        renderer.connect("toggled", self.on_entry_toggled)
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, "active", 2)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        self.filelist.append_column(col)

    # ---------
    # Callbacks
    # ---------

    def on_entry_toggled(self, renderer, path):
        model = self.filelist.get_model()
        iter = model.get_iter(path)
        model.set_value(iter, 2, not model.get_value(iter, 2))

    # --------------
    # Public methods
    # --------------

    def fill_filelist(self):
        """Populate the list of file names from the list of changed editors."""
        model = self.filelist.get_model()
        for i in range(len(self.changed_editors)):
            editor = self.changed_editors[i]
            if editor.get_filename():
                lbl = os.path.basename(editor.get_filename())
                tip = None
            else:
                lbl = _(u"Unsaved Query %(num)s") % {"num" : i+1}
                content = list()
                lines = editor.get_text().splitlines()
                for line in lines:
                    if len(line) > 35:
                        content.append(line[:35]+"...")
                    else:
                        content.append(line)
                    if len(content) == 7:
                        if len(lines) > 7:
                            content.append("...")
                        break
                tip = "\n".join(content)
            model.append([editor, lbl, True, tip])

    def set_dialog_mode(self, mode):
        """Adjust visibility of widgets according to *mode*.

        Usually there's no need to call this method directly. It's called
        on instance creation with the correct mode determined by the list
        of changed editors.

        :param mode: Either *MODE_SINGLE* if just one editor has changed, or
          *MODE_MULTIPLE* if two or more editors have changed.
        """
        if mode == MODE_SINGLE:
            self.builder.get_object("lbl_filelist").hide()
            self.builder.get_object("sw_filelist").hide()
            self.builder.get_object("filelist").hide()
            editor = self.changed_editors[0]
            if editor.get_filename():
                filename = os.path.basename(editor.get_filename())
            else:
                filename = _(u"Unsaved Query %(num)s") % {"num" : 1}
            msg_header = _(u'Save the changes to document "%(filename)s" before closing?') % {"filename":filename}
        else:
            self.builder.get_object("lbl_filelist").show()
            self.builder.get_object("sw_filelist").show()
            msg_header = _(u"There are %(num)d documents with unsaved changes. Save changes before closing?")
            msg_header = msg_header % {"num" : len(self.changed_editors)}
        markup = '<span size="large"><b>%s</b></span>' % msg_header
        self.builder.get_object("lbl_header").set_markup(markup)

    def save_files(self):
        """Save selected files.

        ``True`` is returned if all files were saved. Otherwise ``False``.
        """
        model = self.filelist.get_model()
        iter_ = model.get_iter_first()
        while iter_:
            if model.get_value(iter_, 2):
                editor = model.get_value(iter_, 0)
                default_name = model.get_value(iter_, 1)
                if not editor.save_file(default_name=default_name):
                    return False
            iter_ = model.iter_next(iter_)
        return True

    def run(self):
        """Run the dialog."""
        return self.dlg.run()

    def destroy(self):
        """Destroy the dialog."""
        self.dlg.destroy()

    def hide(self):
        """Hide the dialog."""
        self.dlg.hide()



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SQL editor and results view"""

from gettext import gettext as _
import logging
import os
import re
import string
import thread
import time
import urlparse

import gobject
import gtk
import gtksourceview2
import pango

from cf.db import Query
from cf.plugins.core import PLUGIN_TYPE_EXPORT
from cf.ui import dialogs
from cf.ui.confirmsave import ConfirmSaveDialog
from cf.ui.pane import PaneItem
from cf.ui.widgets import DataExportDialog
from cf.ui.widgets.grid import Grid
from cf.ui.widgets.sqlview import SQLView
from cf.utils import to_uri

import sqlparse


FORMATTER_DEFAULT_OPTIONS = {
    'reindent': True,
    'n_indents': 4,
    'keyword_case': 'upper',
    'identifier_case': 'lower',
#    'right_margin': 79
}


class Editor(gobject.GObject, PaneItem):
    """SQL editor widget.

    :Signals:

    connection-changed
      ``def callback(editor, connection)``

      Emitted when a connection was assigned to this editor.
    """

    name = _(u'Editor')
    icon = gtk.STOCK_EDIT
    detachable = True

    __gsignals__ = {
        "connection-changed" : (gobject.SIGNAL_RUN_LAST,
                                gobject.TYPE_NONE,
                                (gobject.TYPE_PYOBJECT,))
    }

    __gproperties__ = {
        'buffer-dirty': (gobject.TYPE_BOOLEAN,
                         'dirty flag',
                         'is the buffer dirty?',
                         False,
                         gobject.PARAM_READWRITE),
        }

    def __init__(self, win):
        """Constructor.

        :param win: A MainWindow instance.
        """
        self.__gobject_init__()
        PaneItem.__init__(self, win.app)
        self.app = win.app
        self.win = win
        self.connection = None
        self._buffer_dirty = False
        self.__conn_close_tag = None
        self._query_timer = None
        self._filename = None
        self._filecontent_read = ""
        self.builder = gtk.Builder()
        self.builder.set_translation_domain('crunchyfrog')
        self.builder.add_from_file(self.app.get_glade_file('editor.glade'))
        self.widget = self.builder.get_object('box_editor')
        self._setup_widget()
        self._setup_connections()
        self.on_messages_copy = self.results.on_messages_copy
        self.on_messages_clear = self.results.on_messages_clear
        self.on_copy_data = self.results.on_copy_data
        self.on_export_data = self.results.on_export_data
        self.builder.connect_signals(self)
        self.set_data("win", None)
        self.win.emit('editor-created', self)
        self.show_all()
        if self.app.config.get('editor.hide_results_pane'):
            gobject.idle_add(self.toggle_results_pane)

    def show_all(self):
        self.widget.show_all()

    def show(self):
        self.widget.show()

    def destroy(self):
        self.widget.destroy()

    def get_widget(self):
        return self.widget

    def do_get_property(self, param):
        if param.name == 'buffer-dirty':
            return self._buffer_dirty

    def do_set_property(self, param, value):
        if param.name == 'buffer-dirty':
            self._buffer_dirty = value

    # Widget setup

    def _setup_widget(self):
        self._setup_textview()
        self._setup_resultsgrid()

    def _setup_textview(self):
        self.textview = SQLView(self.win, self)
        sw = self.builder.get_object("sw_editor_textview")
        sw.add(self.textview)
        buffer_ = self.textview.buffer
        tag = buffer_.create_tag('error', underline=pango.UNDERLINE_ERROR)
        self.textview.buffer.connect('changed', self.on_buffer_changed)

    def _setup_resultsgrid(self):
        self.results = ResultsView(self.win, self.builder)

    def _setup_connections(self):
        self.textview.connect("populate-popup", self.on_populate_popup)

    # Callbacks

    def on_buffer_changed(self, buffer):
        self.props.buffer_dirty = self.contents_changed()
        iter1 = buffer.get_iter_at_mark(buffer.get_insert())
        iter1.set_line_offset(0)
        iter2 = iter1.copy()
        iter2.forward_to_line_end()
        buffer.remove_tag_by_name('error', iter1, iter2)

    def on_close(self, *args):
        self.close()

    def on_connection_closed(self, connection):
        if connection == self.connection and self.__conn_close_tag:
            connection.disconnect(self.__conn_close_tag)
        self.set_connection(None)

    def on_explain(self, *args):
        self.explain()

    def on_populate_popup(self, textview, popup):
        cfg = self.app.config
        sep = gtk.SeparatorMenuItem()
        sep.show()
        popup.append(sep)
        item = gtk.CheckMenuItem(_(u"Split statements"))
        item.set_active(cfg.get("sqlparse.enabled"))
        item.connect("toggled", lambda x: cfg.set("sqlparse.enabled",
                                                  x.get_active()))
        item.show()
        popup.append(item)
        item = gtk.ImageMenuItem("gtk-close")
        item.show()
        item.connect("activate", self.on_close)
        popup.append(item)

    def on_query_started(self, query):
        start = time.time()
        # Note: The higher the time out value, the longer the query takes.
        #    50 is a reasonable value anyway.
        #    The start time is for the UI only. The real execution time
        #    is calculated in the Query class.
        self.win.statusbar.pop(1)
        if self._query_timer is not None:
            gobject.source_remove(self._query_timer)
        self.results.add_separator()
        self.results.add_message(query.statement, type_='query')
        query.path_status = self.results.add_message("")
        self._query_timer = gobject.timeout_add(50, self.update_exectime,
                                                start, query)

    def on_query_finished(self, query, tag_notice):
        if self._query_timer:
            gobject.source_remove(self._query_timer)
            self._query_timer = None
        self.results.set_query(query)
        if query.failed:
            msg = _(u'Query failed (%(sec).3f seconds)')
            msg = msg % {"sec": query.execution_time}
            type_ = 'error'
            if query.error_position:
                line, offset = query.error_position
                line += query.get_data('editor_start_line')
                self._mark_error(line, offset)
        elif query.description:
            msg = (_(u"Query finished (%(sec).3f seconds, %(num)d rows)")
                   % {"sec": query.execution_time,
                      "num": query.rowcount})
            type_ = 'info'
        else:
            msg = _(u"Query finished (%(sec).3f seconds, "
                    u"%(num)d affected rows)")
            msg = msg % {"sec": query.execution_time,
                         "num": query.rowcount}
            type_ = 'info'
        self.results.add_message(msg, type_, query.path_status)
        self.win.statusbar.push(1, msg)
        if self.connection.handler_is_connected(tag_notice):
            self.connection.disconnect(tag_notice)
        self.textview.grab_focus()

    def on_show_in_main_window(self, *args):
        gobject.idle_add(self.show_in_main_window)

    def on_show_in_separate_window(self, *args):
        gobject.idle_add(self.show_in_separate_window)

    # Public methods

    def _mark_error(self, line, offset):
        line = line - 2
        offset = offset - 1
        if line < 0 or offset < 0:
            # Backend reported a bad error position. Simply ignore it.
            return
        buffer_ = self.textview.buffer
#        buffer_.create_source_mark(
#            None, 'error',
#            buffer_.get_iter_at_line_offset(line-2, offset))
        tag_table = buffer_.get_tag_table()
        tag = tag_table.lookup('error')
        iter_err = buffer_.get_iter_at_line_offset(line, offset)
        iter1 = iter_err.copy()
        if not iter1.starts_word():
            iter1.backward_word_start()
        iter2 = iter1.copy()
        iter2.forward_word_end()
#        tag = gtk.TextTag()
#        tag.props.underline = pango.UNDERLINE_SINGLE
        buffer_.apply_tag(tag, iter1, iter2)
        buffer_.place_cursor(iter_err)
        self.textview.move_mark_onscreen(buffer_.get_insert())
#        self.textview.set_mark_category_background('error',
#                                                   gtk.gdk.color_parse('red'))
#        it = gtk.icon_theme_get_default()
#        pb = it.load_icon ("gtk-dialog-error", gtk.ICON_SIZE_MENU,
#                           gtk.ICON_LOOKUP_USE_BUILTIN)
#        self.textview.set_mark_category_pixbuf('error', pb)
#        self.textview.set_show_line_marks(True)

    def get_focus_child(self):
        return self.get_child1().get_children()[0].grab_focus()

    def close(self, force=False):
        """Close editor, displays a confirmation dialog for unsaved files.

        Args:
          force: If True, the method doesn't check for changed contents.
        """
        if self.contents_changed() and not force:
            dlg = ConfirmSaveDialog(self.win, [self])
            resp = dlg.run()
            if resp == 1:
                ret = dlg.save_files()
            elif resp == 2:
                ret = True
            else:
                ret = False
            dlg.destroy()
        else:
            ret = True
        if ret:
            if self.get_data("win"):
                self.get_data("win").destroy()
            else:
                self.destroy()
            self.win.set_editor_active(self, False)
            self.win.editor_remove(self)
            return True

    def commit(self):
        """Commit current transaction, if any."""
        if not self.connection: return
        self.connection.commit()
        self.results.add_message('COMMIT', 'info')

    def rollback(self):
        """Commit current transaction, if any."""
        if not self.connection: return
        self.connection.rollback()
        self.results.add_message('ROLLBACK', 'info')

    def begin_transaction(self):
        """Begin transaction."""
        if not self.connection: return
        self.connection.begin()
        self.results.add_message('BEGIN TRANSACTION', 'info')

    def execute_query(self, statement_at_cursor=False):
        # TODO(andi): This method needs some refactoring:
        #   - the actual execution code is doubled
        #   - that affects offset counting too
        self.results.assure_visible()
        def exec_threaded(statement, start_line):
            if self.app.config.get("sqlparse.enabled", True):
                stmts = sqlparse.split(statement)
            else:
                stmts = [statement]
            for stmt in stmts:
                add_offset = len(stmt.splitlines())
                if not stmt.strip():
                    start_line += add_offset
                    continue
                query = Query(stmt, self.connection)
#                query.coding_hint = self.connection.coding_hint
                gtk.gdk.threads_enter()
                query.set_data('editor_start_line', start_line)
                query.connect("started", self.on_query_started)
                query.connect("finished",
                              self.on_query_finished,
                              tag_notice)
                gtk.gdk.threads_leave()
                query.execute(True)
                start_line += add_offset
                if query.failed:
                    # hmpf, doesn't work that way... so just return here...
                    return
#                    gtk.gdk.threads_enter()
#                    dlg = gtk.MessageDialog(None,
#                                            gtk.DIALOG_MODAL|
#                                            gtk.DIALOG_DESTROY_WITH_PARENT,
#                                            gtk.MESSAGE_ERROR,
#                                            gtk.BUTTONS_YES_NO,
#                                            _(u"An error occurred. Continue?"))
#                    if dlg.run() == gtk.RESPONSE_NO:
#                        leave = True
#                    else:
#                        leave = False
#                    dlg.destroy()
#                    gtk.gdk.threads_leave()
#                    if leave:
#                        return
        buffer = self.textview.get_buffer()
        self.results.reset()
        if not statement_at_cursor:
            bounds = buffer.get_selection_bounds()
            if not bounds:
                bounds = buffer.get_bounds()
        else:
            bounds = self.textview.get_current_statement()
            if bounds is None:
                return
        buffer.remove_tag_by_name('error', *bounds)
        statement = buffer.get_text(*bounds)
        if self.app.config.get("editor.replace_variables"):
            tpl = string.Template(statement)
            tpl_search = tpl.pattern.search(tpl.template)
            if tpl_search and tpl_search.groupdict().get("named"):
                dlg = StatementVariablesDialog(tpl)
                if dlg.run() == gtk.RESPONSE_OK:
                    statement = dlg.get_statement()
                else:
                    statement = None
                dlg.destroy()
                if not statement:
                    return
        def foo(connection, msg):
            self.results.add_message(msg)
        tag_notice = self.connection.connect("notice", foo)
        if self.connection.threadsafety >= 2:
            start_line = bounds[0].get_line()+1
            thread.start_new_thread(exec_threaded, (statement, start_line))
        else:
            start_line = bounds[0].get_line()+1
            line_offset = start_line
            if self.app.config.get("sqlparse.enabled", True):
                stmts = sqlparse.split(statement)
            else:
                stmts = [statement]
            for stmt in stmts:
                add_offset = len(stmt.splitlines())
                if not stmt.strip():
                    line_offset += add_offset
                    continue
                query = Query(stmt, self.connection)
                query.set_data('editor_start_line', line_offset)
#                query.coding_hint = self.connection.coding_hint
                query.connect("started", self.on_query_started)
                query.connect("finished", self.on_query_finished, tag_notice)
                query.execute()
                line_offset += add_offset

    def explain(self):
        self.results.assure_visible()
        buf = self.textview.get_buffer()
        bounds = buf.get_selection_bounds()
        if not bounds:
            bounds = buf.get_bounds()
        statement = buf.get_text(*bounds)
        if len(sqlparse.split(statement)) > 1:
            dialogs.error(_(u"Select a single statement to explain."))
            return
        if not self.connection:
            return
        queries = [Query(stmt, self.connection)
                   for stmt in self.connection.explain_statements(statement)]
        def _execute_next(last, queries):
            if last is not None and last.failed:
                self.results.set_explain_results(last)
                return
            q = queries.pop(0)
            if len(queries) == 0:
                q.connect('finished',
                          lambda x: self.results.set_explain_results(x))
            else:
                q.connect('finished',
                          lambda x: _execute_next(x, queries))
            q.execute()
        _execute_next(None, queries)

    def set_connection(self, conn):
        if self.connection and self.__conn_close_tag:
            if self.connection.handler_is_connected(self.__conn_close_tag):
                self.connection.disconnect(self.__conn_close_tag)
            self.__conn_close_tag = None
        self.connection = conn
        if conn:
            self.__conn_close_tag = self.connection.connect("closed",
                                                            self.on_connection_closed)
        else:
            self._conn_close_tag = None
        self.emit("connection-changed", conn)

    def get_connection(self):
        """Returns the connection assigned to the editor."""
        return self.connection

    def set_filename(self, filename):
        """Opens filename.

        Returns ``True`` if the file was successfully opened.
        Otherwise ``False``.
        """
        msg = None
        if not os.path.isfile(filename):
            msg = _(u'No such file: %(name)s')
        elif not os.access(filename, os.R_OK):
            msg = _(u'File is not readable: %(name)s')
        if msg is not None:
            dialogs.error(_(u"Failed to open file"), msg % {'name': filename})
            return False
        self._filename = filename
        if filename:
            f = open(self._filename)
            a = f.read()
            f.close()
        else:
            a = ""
        self._filecontent_read = a
        self.set_text(a)
        self.app.recent_manager.add_item(to_uri(filename))
        return True

    def get_filename(self):
        return self._filename

    def file_contents_changed(self):
        if self._filename:
            buffer = self.textview.get_buffer()
            return buffer.get_text(*buffer.get_bounds()) != self._filecontent_read
        return False

    def contents_changed(self):
        if self._filename:
            return self.file_contents_changed()
        elif len(self.get_text()) == 0:
            return False
        return True

    def file_confirm_save(self):
        dlg = dialogs.yesno(_(u"Save file %(name)s before closing the editor?") % {"name":os.path.basename(self._filename)})
        if dlg == gtk.RESPONSE_YES:
            return self.save_file_as()
        return True

    def save_file(self, parent=None, default_name=None):
        if not self._filename:
            return self.save_file_as(parent=parent, default_name=default_name)
        buffer = self.get_buffer()
        a = buffer.get_text(*buffer.get_bounds())
        f = open(self._filename, "w")
        f.write(a)
        f.close()
        self.app.recent_manager.add_item(to_uri(self._filename))
        self._filecontent_read = a
        gobject.idle_add(buffer.emit, "changed")
        return True

    def save_file_as(self, parent=None, default_name=None):
        if not parent:
            parent = self.win
        dlg = gtk.FileChooserDialog(_(u"Save file"),
                            parent,
                            gtk.FILE_CHOOSER_ACTION_SAVE,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        if self._filename:
            dlg.set_filename(self._filename)
        else:
            dlg.set_current_folder(self.app.config.get("editor.recent_folder", ""))
            if default_name:
                dlg.set_current_name(default_name)
        filter = gtk.FileFilter()
        filter.set_name(_(u"All files (*)"))
        filter.add_pattern("*")
        dlg.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name(_(u"SQL files (*.sql)"))
        filter.add_pattern("*.sql")
        dlg.add_filter(filter)
        dlg.set_filter(filter)
        if dlg.run() == gtk.RESPONSE_OK:
            self._filename = dlg.get_filename()
            self.save_file()
            self.app.config.set("editor.recent_folder", dlg.get_current_folder())
            ret = True
        else:
            ret = False
        dlg.destroy()
        return ret

    def get_buffer(self):
        return self.textview.get_buffer()

    def get_text(self):
        buffer = self.get_buffer()
        return buffer.get_text(*buffer.get_bounds())

    def set_text(self, txt):
        buffer_ = self.get_buffer()
        buffer_.begin_user_action()
        buffer_.set_text(txt)
        buffer_.end_user_action()  # starts tagging of statements too

    def show_in_separate_window(self):
        instance = self.app.new_instance(show=False)
        instance.editor_append(self)
        instance.show()
    detach = show_in_separate_window  # make it compatible with PaneItem

    def show_in_main_window(self):
        self.win.queries.attach(self)
        win = self.get_data("win")
        if win:
            win.destroy()
        self.set_data("win", None)

    def update_exectime(self, start, query):
        lbl = _("Query running... (%.3f seconds)" % (time.time()-start))
        self.results.add_message(lbl, path=query.path_status)
        if query.executed:
            if self._query_timer is not None:
                gobject.source_remove(self._query_timer)
            self._query_timer = None
            return False
        else:
            return True

    # Printing

    def on_print_paginate(self, operation, context, compositor):
        if compositor.paginate(context):
            n_pages = compositor.get_n_pages()
            operation.set_n_pages(n_pages)
            return True
        return False

    def on_print_draw_page(self, operation, context, page_no, compositor):
        compositor.draw_page(context, page_no)

    def on_end_page(self, operation, context, compositor):
        pass

    def print_contents(self, preview=False):
        """Send content of editor to printer."""
        view = self.textview
        compositor = gtksourceview2.print_compositor_new_from_view(view)
        operation = gtk.PrintOperation()
        operation.connect('paginate', self.on_print_paginate, compositor)
        operation.connect('draw-page', self.on_print_draw_page, compositor)
        if preview:
            action = gtk.PRINT_OPERATION_ACTION_PREVIEW
        else:
            action = gtk.PRINT_OPERATION_ACTION_PRINT_DIALOG
        operation.run(action)

    # Clipboard functions

    def clipboard_copy(self, clipboard):
        """Copies selected data to clipboard.

        This is either the selected text from the editor or the selected
        cells from the results grid.
        """
        if self.textview.is_focus():
            buffer_ = self.textview.get_buffer()
            buffer_.copy_clipboard(clipboard)
        elif self.results.grid.grid.is_focus():
            self.results.clipboard_copy(clipboard)

    def clipboard_cut(self, clipboard):
        """Cuts selected text from editor."""
        buffer_ = self.textview.get_buffer()
        buffer_.cut_clipboard(clipboard, True)

    def clipboard_paste(self, clipboard):
        """Pastes clipboard data to editor."""
        buffer_ = self.textview.get_buffer()
        buffer_.paste_clipboard(clipboard, None, True)

    # Statement navigation and formatting

    def rjump_to_statement(self, offset):
        """Jumps to a statement relative to current.

        :param offset: Position relative to current statement as integer.
        """
        idx_current = 0
        positions = []
        curr = self.textview.get_current_statement()
        buffer_ = self.textview.get_buffer()
        if curr is not None:
            # We're using the presence of an end iter as a flag if we're
            # inside a statement. If not, just jump to the next/prev statement
            # without respecting the offset.
            cstart, inside_statement = curr
        else:
            cstart = buffer_.get_iter_at_mark(buffer_.get_insert())
            cstart.set_line_offset(0)
            inside_statement = None
        for start, end in self.textview.get_statements():
            positions.append((start, end))
            if not inside_statement:
                if offset > 0 and start.get_line() > cstart.get_line():
                    buffer_.place_cursor(start)
                    return
                elif offset < 0 and start.get_line() > cstart.get_line():
                    if len(positions) > 2:
                        buffer_.place_cursor(positions[-2][0])
                    elif positions:  # there's just one statement buffered
                        buffer_.place_cursor(positions[0][0])
                    return
            elif start.equal(cstart):
                idx_current = len(positions)-1
        max_idx = len(positions)-1
        if offset > 0:
            new_pos = min(idx_current+offset, max_idx)
        else:
            new_pos = max(idx_current+offset, 0)
        buffer_.place_cursor(positions[new_pos][0])

    def selected_lines_toggle_comment(self):
        """Comments/uncomments selected lines."""
        buffer_ = self.textview.get_buffer()
        res = buffer_.get_selection_bounds()
        if not res:
            start = buffer_.get_iter_at_mark(buffer_.get_insert())
            lno_start = lno_end = start.get_line()
        else:
            start, end = res
            lno_start = start.get_line()
            lno_end = end.get_line()
        buffer_.begin_user_action()
        for line_no in xrange(lno_start, lno_end+1):
            lstart = buffer_.get_iter_at_line(line_no)
            lend = lstart.copy()
            lend.forward_to_line_end()
            line = buffer_.get_text(lstart, lend)
            if not line.startswith('-- '):
                line = '-- %s' % line
            else:
                line = re.sub(r'^\s*-- ', '', line)
            buffer_.delete(lstart, lend)
            buffer_.insert(lstart, line)
        buffer_.end_user_action()

    def selected_lines_quick_format(self, **options):
        """Runs format without any other options than the default ones."""
        if not options:
            options = FORMATTER_DEFAULT_OPTIONS
        buffer_ = self.textview.get_buffer()
        res = buffer_.get_selection_bounds()
        if not res:
            start = end = None
            if self.app.config.get('editor.format_statement_at_cursor', False):
                curr = self.textview.get_current_statement()
                if curr:
                    start, end = curr
                    select_range = False
            if start is None:
                start, end = buffer_.get_bounds()
                select_range = False
        else:
            start, end = res
            select_range = True
        # Count chars excluding whitespaces
        insert_mark = buffer_.get_insert()
        insert_iter = buffer_.get_iter_at_mark(insert_mark)
        if insert_iter.in_range(start, end):
            char_offset = len(re.sub(r'\s', '',
                                     buffer_.get_text(start, insert_iter)))
            start_offset = start.get_offset()
        else:
            char_offset = start_offset = None
        orig = buffer_.get_text(start, end)
        formatted = sqlparse.format(orig, **options)
        # Modify buffer
        buffer_.begin_user_action()
        buffer_.delete(start, end)
        buffer_.insert(start, formatted)
        # Set selection again
        if select_range:
            end = start.copy()
            end.backward_chars(len(formatted))
            buffer_.select_range(end, start)
        buffer_.end_user_action()
        # TODO: place cursor - but how to do this... :)
        if char_offset is not None:
            num = 0
            idx = 0
            iter_ = buffer_.get_iter_at_offset(start_offset)
            for i in range(len(formatted)):
                if formatted[i] not in '\r\n\t ':
                    num += 1
                if num == char_offset:
                    iter_.forward_chars(i+1)
                    buffer_.place_cursor(iter_)
                    self.textview.scroll_to_mark(buffer_.get_insert(), 0.25)
                    break

    def toggle_results_pane(self):
        """Show or hide the result pane."""
        pane = self.widget.get_child2()
        if pane.get_property('visible'):
            pane.hide()
        else:
            pane.show()


class ResultsView(object):

    def __init__(self, win, builder):
        self.instance = win
        self.app = win.app
        self.widget = builder.get_object('editor_results')
        self.builder = builder
        self._setup_widget()
        self._setup_connections()

    def _setup_widget(self):
        self.grid = ResultList(self.instance, self.builder)
        self.messages = self.builder.get_object("editor_results_messages")
        model = gtk.ListStore(str,  # 0 stock id
                              str,  # 1 message
                              str,  # 2 foreground color
                              int,  # 3 font weight
                              bool, # 4 is separator row
                              str,  # 5 font description
                              )
        model.connect("row-inserted", self._msg_model_changed)
        model.connect("row-deleted", self._msg_model_changed)
        self.messages.set_model(model)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'stock-id', 0)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        col.add_attribute(renderer, 'foreground', 2)
        col.add_attribute(renderer, 'weight', 3)
        col.add_attribute(renderer, 'font', 5)
        self.messages.append_column(col)
        self.messages.set_row_separator_func(self._set_row_separator)
        self.explain_results = Grid()
        sw = self.builder.get_object('sw_explain_results')
        sw.add(self.explain_results)
        self.explain_results.show_all()
        self._update_btn_export_state()

    def _update_btn_export_state(self):
        """Update state of export button."""
        btn_export = self.builder.get_object('editor_export_data')
        rows = None
        if self.grid.query:
            rows = self.grid.query.rows
        sensitive = False
        if rows:
            sensitive = True
            exp_plugins = self.app.plugins.get_plugins(PLUGIN_TYPE_EXPORT,
                                                       True)
            if not exp_plugins:
                sensitive = False
        btn_export.set_sensitive(sensitive)

    def _msg_model_changed(self, model, *args):
        tb_clear = self.builder.get_object("tb_messages_clear")
        tb_copy = self.builder.get_object("tb_messages_copy")
        sensitive = model.get_iter_first() is not None
        tb_clear.set_sensitive(sensitive)
        tb_copy.set_sensitive(sensitive)

    def _setup_connections(self):
        self.grid.grid.connect("selection-changed",
                               self.on_grid_selection_changed)
        self.app.plugins.connect('plugin-added',
                                 lambda *a: self._update_btn_export_state())
        self.app.plugins.connect('plugin-removed',
                                 lambda *a: self._update_btn_export_state())
        self.app.plugins.connect('plugin-active',
                                 lambda *a: self._update_btn_export_state())

    def _set_row_separator(self, model, iter):
        return model.get_value(iter, 4)

    def on_copy_data(self, *args):
        gobject.idle_add(self.copy_data)

    def on_export_data(self, *args):
        gobject.idle_add(self.export_data)

    def on_messages_clear(self, *args):
        self.messages.get_model().clear()

    def on_messages_copy(self, *args):
        model = self.messages.get_model()
        iter_ = model.get_iter_first()
        plain = []
        while iter_ is not None:
            if model.get_value(iter_, 4): # Is it a separator?
                plain.append('-'*20)
            else:
                value = model.get_value(iter_, 1)
                if value is not None:
                    plain.append(value)
            iter_ = model.iter_next(iter_)
        clipboard = gtk.clipboard_get()
        clipboard.set_text('\n'.join(plain))

    def on_grid_selection_changed(self, grid, selected_cells):
        self.builder.get_object("editor_copy_data").set_sensitive(bool(selected_cells))

    def assure_visible(self):
        """Make sure that the result pane is visible."""
        pane = self.widget.get_parent()
        if not pane.get_property('visible'):
            pane.show()

    def copy_data(self, clipboard=None):
        """Copy selected values to clipboard.

        Columns are terminated by \t, rows by \n.
        If *clipboard* is ``None``, the default clipboard will be used.
        """
        rows = {}
        for cell in self.grid.grid.get_selected_cells():
            value = self.grid.grid.get_cell_data(cell, repr=True)
            if rows.has_key(cell[0]):
                rows[cell[0]] += "\t%s" % value
            else:
                rows[cell[0]] = "%s" % value
        rownums = rows.keys()
        rownums.sort()
        txt = "\n".join(rows.get(rownum) for rownum in rownums)
        if clipboard is None:
            display = gtk.gdk.display_manager_get().get_default_display()
            clipboard = gtk.Clipboard(display, "CLIPBOARD")
        clipboard.set_text(txt)
        self.instance.statusbar.set_message(_(u"Data copied to clipboard"))

    def export_data(self):
        data = self.grid.grid.get_grid_data()
        description = self.grid.grid.description
        selected = self.grid.grid.get_selected_rows()
        statement = self.grid.query.statement
        gtk.gdk.threads_enter()
        dlg = DataExportDialog(self.instance.app, self.instance,
                               data, selected, statement, description)
        if dlg.run() == gtk.RESPONSE_OK:
            dlg.hide()
            dlg.export_data()
        dlg.destroy()
        gtk.gdk.threads_leave()

    def clipboard_copy(self, clipboard):
        """Copies selected cells to clipboard."""
        # TODO: Cleanup API. There should be one function to do this.
        self.copy_data(clipboard)

    def reset(self):
        # Explain
        self.explain_results.reset()
        # Messages
        # Only gray out messages from previous runs, don't remove them!
        model = self.messages.get_model()
        iter_ = model.get_iter_first()
        while iter_:
            model.set_value(iter_, 2, '#cccccc')
            iter_ = model.iter_next(iter_)
        self.widget.set_current_page(2)

    def set_explain_results(self, query):
        self.explain_results.reset()
        if query.failed:
            dialogs.error(_(u'Failed'), '\n'.join(query.errors),
                          parent=self.instance)
        else:
            self.explain_results.set_result(query.rows, query.description)

    def set_query(self, query):
        self.assure_visible()
        self.grid.set_query(query)
        model = self.messages.get_model()
        for err in query.errors:
            self.add_error(err.strip(), monospaced=True)
        for msg in query.messages:
            self.add_output(msg.strip())
        if query.errors:
            curr_page = 2
        elif query.description:
            curr_page = 0
        else:
            curr_page = 2
        self._update_btn_export_state()
        gobject.idle_add(self.widget.set_current_page, curr_page)

    def add_message(self, msg, type_=None, path=None, monospaced=False):
        """Add a message.

        Args:
          msg: The message to add.
          type_: Message type ('info', 'output',
                 'error', 'warning', 'query', None).
        """
        self.assure_visible()
        assert type_ in (None, 'info', 'output', 'error', 'warning', 'query')
        stock_id = None
        foreground = None
        if monospaced:
            font = 'Monospace 10'
        else:
            font = None
        weight = pango.WEIGHT_NORMAL
        if type_ == 'info':
            stock_id = 'gtk-info'
            foreground = '#336699'
        elif type_ == 'error':
            stock_id = 'gtk-dialog-error'
            foreground = '#a40000'
            weight = pango.WEIGHT_BOLD
        elif type_ == 'warning':
            stock_id = 'gtk-dialog-warning'
            foreground = '#00a400'
        elif type_ == 'query':
            stock_id ='gtk-execute'
            weight = pango.WEIGHT_BOLD
        elif type_ == 'output':
            stock_id = 'gtk-go-back'
            font = 'Monospace'
        model = self.messages.get_model()
        if model is None:  # we are in shutdown phase
            return
        msg = msg.strip()
        if path is None:
            itr = model.append([stock_id, msg, foreground, weight,
                                False, font])
        else:
            try:
                iter_ = model.get_iter(path)
            except ValueError:
                return
            model.set_value(iter_, 0, stock_id)
            model.set_value(iter_, 1, msg)
            model.set_value(iter_, 2, foreground)
            model.set_value(iter_, 3, weight)
            model.set_value(iter_, 5, font)
            itr = iter_
        # Somehow this works smarter than treeview.scroll_to_cell(path).
        # See: http://www.mail-archive.com/pygtk@daa.com.au/msg17059.html
        self.messages.scroll_to_cell(str(len(model)-1))
        self.widget.set_current_page(2)
        return model.get_path(itr)

    def add_error(self, msg, monospaced=False):
        return self.add_message(msg, 'error', monospaced=monospaced)

    def add_info(self, msg):
        return self.add_message(msg, 'info')

    def add_warning(self, msg):
        return self.add_message(msg, 'warning')

    def add_output(self, msg):
        return self.add_message(msg, 'output')

    def add_separator(self):
        model = self.messages.get_model()
        if not model.get_iter_first():
            return
        model.append([None, None, None, pango.WEIGHT_NORMAL, True, None])


class ResultList(object):
    """Result list with toolbar"""

    def __init__(self, win, builder):
        self.builder = builder
        self.widget = self.builder.get_object('editor_results_data')
        self._setup_widget()
        self.instance = win
        self.query = None

    def _setup_widget(self):
        self.grid = Grid()
        self.builder.get_object("sw_grid").add(self.grid)

    def set_query(self, query):
        self.query = query
        self.grid.reset()
        if self.query.description:
            try:
                self.grid.set_result(self.query.rows, self.query.description,
                                     self.query.coding_hint)
            except Exception, err:
                logging.exception('Failed to display query results')
                dialogs.error(_(u'Failed to display results'), str(err))



class StatementVariablesDialog(gtk.Dialog):

    def __init__(self, template):
        gtk.Dialog.__init__(self, _(u"Variables"),
                            None,
                            gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OK, gtk.RESPONSE_OK))
        self.template = template
        self._widgets = dict()
        self._setup_widget()

    def _setup_widget(self):
        vars = []
        found = []
        for match in self.template.pattern.finditer(self.template.template):
            name = match.groupdict().get("named")
            if not name or name.lower() in found:
                continue
            vars.append(name)
            found.append(name.lower())
        table = gtk.Table(len(vars), 2)
        table.set_row_spacings(5)
        table.set_col_spacings(7)
        for i in range(len(vars)):
            lbl = gtk.Label(vars[i])
            lbl.set_alignment(0, 0.5)
            table.attach(lbl, 0, 1, i, i+1, gtk.FILL, gtk.FILL)
            entry = gtk.Entry()
            table.attach(entry, 1, 2, i, i+1, gtk.EXPAND|gtk.FILL, gtk.FILL)
            self._widgets[vars[i]] = entry
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add_with_viewport(table)
        sw.set_border_width(10)
        self.vbox.pack_start(sw, True, True)
        self.vbox.show_all()

    def on_value_edited(self, renderer, path, value):
        model = self.treeview.get_model()
        iter = model.get_iter(path)
        model.set_value(iter, 1, value)

    def get_statement(self):
        data = dict()
        for var, widget in self._widgets.items():
            data[var] = widget.get_text()
        return self.template.safe_substitute(data)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SQL Server backend."""

from gettext import gettext as _

from cf.db import TRANSACTION_ACTIVE, TRANSACTION_IDLE
from cf.db import objects
from cf.db.backends import Generic, DEFAULT_OPTIONS, GUIOption


class MSSql(Generic):

    drivername = 'mssql'

    @classmethod
    def dbapi(cls):
        import pymssql
        return pymssql

    @classmethod
    def get_connect_params(cls, url):
        opts = url.get_dict()
        if 'username' in opts:
            opts['user'] = opts['username']
            del opts['username']
        if 'port' in opts:
            opts['host'] = '%s:%s' % (opts['host'], opts['port'])
            del opts['port']
        return tuple(), opts

    @classmethod
    def get_options(cls):
        opts = list(DEFAULT_OPTIONS)
        opts.append(
            GUIOption(
                'charset', _('Character set'),
                tooltip=_('Character set with which to connect to the database')
                ))
        return tuple(opts)

    def get_server_info(self, connection):
        ver = connection.execute('SELECT @@VERSION')[0][0]
        ver = ver.splitlines()[0].strip()+" - "+ver.splitlines()[-1].strip()
        return ver

    def begin(self, connection):
        connection.execute('BEGIN TRANSACTION')
        connection._tstate = TRANSACTION_ACTIVE
        return True

    def commit(self, connection):
        super(MSSql, self).commit(connection)
        connection._tstate = TRANSACTION_IDLE

    def rollback(self, connection):
        super(MSSql, self).rollback(connection)
        connection._tstate = TRANSACTION_IDLE

    def get_transaction_state(self, connection):
        return getattr(connection, '_tstate', TRANSACTION_IDLE)

    def _query(self, connection, sql):
        return connection.execute_raw(sql)

    def initialize(self, meta, connection):
        schemata = objects.Schemata(meta)
        meta.set_object(schemata)
        for item in self._query(connection, INITIAL_SQL):
            schema = meta.find_exact(cls=objects.Schema,
                                     name=item['table_schema'])
            if schema is None:
                schema = objects.Schema(meta, name=item['table_schema'],
                                        parent=schemata)
                meta.set_object(schema)
            if item['table_type'] == 'BASE TABLE':
                cls = objects.Table
                coll_cls = objects.Tables
            elif item['table_type'] == 'VIEW':
                cls = objects.View
                coll_cls = objects.Views
            else:
                continue
            parent = meta.find_exact(cls=coll_cls, parent=schema)
            if parent is None:
                parent = coll_cls(meta, parent=schema)
                meta.set_object(parent)
            obj = meta.find_exact(cls=cls, name=item['table_name'],
                                  parent=parent)
            if obj is None:
                obj = cls(meta, name=item['table_name'],
                          parent=parent, schema=schema)
                meta.set_object(obj)
            col = meta.find_exact(cls=objects.Column, name=item['column_name'],
                                  parent=obj.columns)
            if col is None:
                col = objects.Column(meta, name=item['column_name'],
                                     parent=obj.columns)
                meta.set_object(col)


DRIVER = MSSql


INITIAL_SQL = """SELECT t.table_catalog, t.table_schema,
t.table_name, c.column_name, t.table_type
FROM INFORMATION_SCHEMA.COLUMNS c,
    INFORMATION_SCHEMA.TABLES t
WHERE t.TABLE_NAME = c.TABLE_NAME
      AND t.table_catalog = c.table_catalog
ORDER BY c.table_catalog,
         c.TABLE_NAME,
         c.ordinal_position
"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Python shell plugin"""

import logging
from gettext import gettext as _

import gtk

from cf.plugins.core import BottomPanePlugin
from cf.plugins.mixins import InstanceMixin
from cf.ui.pane import PaneItem

try:
    from ipython_view import IPythonView
    HAVE_IPYTHON = True
except ImportError, err:
    logging.error(err)
    HAVE_IPYTHON = False


class CFShell(BottomPanePlugin, InstanceMixin):

    id = "crunchyfrog.plugin.cfshell"
    name = _(u"Shell")
    description = _(u"Interactive shell (mainly for debugging)")
    icon = "gnome-terminal"
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://cf.andialbrecht.de"
    version = "0.1"

    def __init__(self, app):
        BottomPanePlugin.__init__(self, app)
        self._instances = dict()

    def init_instance(self, instance):
        if instance in self._instances:
            return
        view = CFShellView(self.app, instance)
        instance.bottom_pane.add_item(CFShellView(self.app, instance))
        self._instances[instance] = view

    def shutdown(self):
        while self._instances:
            instance, view = self._instances.popitem()
            view.destroy()

if not HAVE_IPYTHON:
    CFShell.INIT_ERROR = _(u'Python module ipython is required.')


class CFShellView(gtk.ScrolledWindow, PaneItem):

    name = _(u'Shell')
    icon = 'gnome-terminal'
    detachable = True

    def __init__(self, app, instance):
        gtk.ScrolledWindow.__init__(self)
        PaneItem.__init__(self, app)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        self.app = app
        self.instance = instance
        self.iview = IPythonView()
        self.iview.updateNamespace({"app": self.app,
                                    "window": self.instance})
        self.add(self.iview)
        self.set_size_request(-1, 100)
        self.show_all()



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""SQLite backend"""

from gettext import gettext as _

import gobject
import os
import re
import functools

from cf.db import TRANSACTION_IDLE, TRANSACTION_ACTIVE
from cf.db import objects
from cf.db.backends import Generic, GUIOption


P_ERROR_MESSAGES = set([
    re.compile(r'^no such (table|column): (?P<pattern>.*)$'),
    re.compile(r'^near "(?P<pattern>.*)": syntax error$'),
    re.compile(r'table (?P<pattern>.*) already exists$'),
    re.compile(r'there is already another table or index with this name: (?P<pattern>.*)$'),
    ])


class SQLite(Generic):

    drivername = 'sqlite'

    @classmethod
    def get_options(cls):
        return (GUIOption('database', _(u'File'), required=True),)

    @classmethod
    def create_url(cls, options):
        options['database'] = os.path.expanduser(options['database'])
        return super(SQLite, cls).create_url(options)

    @classmethod
    def get_native_shell_command(cls, url):
        return 'sqlite3', [url.database,]

    @classmethod
    def dbapi(cls):
        import sqlite3
        return sqlite3

    @classmethod
    def should_close(cls, error):
        return False

    @classmethod
    def get_error_position(cls, query, error):
        pattern = None
        for p in P_ERROR_MESSAGES:
            m = p.match(str(error))
            if m:
                pattern = m.groupdict()['pattern']
                break
        if pattern is None:
            return None
        for idx, line in enumerate(query.statement.splitlines()):
            if pattern in line:
                return idx+1, line.index(pattern)+1
        return None

    def get_connect_params(self, url):
        return (url.database,), {}

    def prepare_connection(self, connection):
        connection.connection.isolation_level = None
        fpart = functools.partial(self._transaction_watcher, connection)
        connection.connection.set_authorizer(fpart)

    def _transaction_watcher(self, connection, action_code, operation, *args):
        sqlite3 = self.dbapi()
        if action_code != sqlite3.SQLITE_TRANSACTION:
            return sqlite3.SQLITE_OK
        if operation.upper() == 'BEGIN':
            gobject.idle_add(connection.set_property,
                             'transaction-state', TRANSACTION_ACTIVE)
        elif operation.upper() in ('ROLLBACK', 'COMMIT'):
            gobject.idle_add(connection.set_property,
                             'transaction-state', TRANSACTION_IDLE)
            fpart = functools.partial(self._transaction_watcher, connection)
            connection.connection.set_authorizer(fpart)
        return sqlite3.SQLITE_OK

    def get_server_info(self, connection):
        return 'SQLite %s' % self.dbapi().sqlite_version

    def initialize(self, meta, connection):
        sql = 'select type, name, tbl_name, sql from sqlite_master'
        tables = objects.Tables(meta)
        meta.set_object(tables)
        views = objects.Views(meta)
        meta.set_object(views)
        for item in connection.execute(sql):
            if item[0] == 'table':
                table = objects.Table(meta, name=item[1], parent=tables,
                                      createstmt=item[3])
                meta.set_object(table)
            elif item[0] == 'view':
                views = objects.View(meta, name=item[1], parent=views,
                                     createstmt=item[3])
                meta.set_object(views)

    def refresh(self, obj, meta, connection):
        if obj.typeid == 'columns':
            self._refresh_columns(obj, meta, connection)

    def _refresh_columns(self, obj, meta, connection):
        table = obj.parent
        known_columns = {}
        [known_columns.setdefault(k.cid, k)
         for k in meta.find(cls=objects.Column, parent=obj)]
        sql = "pragma table_info('%s')" % table.name  # somehow ? doesn't work
        for item in connection.execute(sql):
            # item is: cid, name, type, notnull, dflt_value, pk
            col = known_columns.get(item[0], None)
            if col is None:
                col = objects.Column(meta, parent=obj, cid=item[0])
            col.name = item[1]
            col.type = item[2]
            col.notnull = item[3]
            col.default = item[4]
            col.pk = item[5]
            meta.set_object(col)


DRIVER = SQLite



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Dialog to manage data sources and connections."""

import sys
from gettext import gettext as _

import gobject
import gtk

import cf.db
from cf.db.backends import GUIOption
from cf.ui import dialogs
from cf.ui.widgets.sqlview import SQLView


class DatasourcesDialog(object):


    def __init__(self, app, parent=None):
        self.app = app
        self.builder = gtk.Builder()
        self.builder.set_translation_domain('crunchyfrog')
        self.builder.add_from_file(app.get_glade_file('datasources.glade'))
        self.builder.connect_signals(self)
        self.dlg = self.builder.get_object('datasources_dialog')
        self.dlg.set_transient_for(parent)
        self.init_datasources()
        self.refresh_datasources()
        self.init_providers()
        self.refresh_providers()
        self._ds_sigs = []
        ds = self.app.datasources
        self._ds_sigs.append(ds.connect('datasource-changed',
                                        lambda *a: self.refresh_datasources()))
        self._ds_sigs.append(ds.connect('datasource-added',
                                        lambda *a: self.refresh_datasources()))
        self._ds_sigs.append(ds.connect('datasource-deleted',
                                        lambda *a: self.refresh_datasources()))

    # -----------------
    # Dialog callbacks
    # -----------------

    def on_close(self, *args):
        self.app.set_data('dialog_datasources', None)
        self.destroy()

    def on_response(self, dialog, response_id):
        if response_id == -1:  # help button
            dialog.stop_emission('response')
            self.app.show_help('datasources')
            return True

    def on_show_help(self, *args):
        self.app.show_help('datasources')
        self.stop_emission('clicked')
        return True

    # -----------
    # Datsources
    # -----------

    def init_datasources(self):
        treeview = self.builder.get_object('list_datasources')
        model = gtk.ListStore(gobject.TYPE_PYOBJECT,  # 0 datasource
                              str,                    # 1 label
                              str,                    # 2 color
                              )
        model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        treeview.set_model(model)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'background', 2)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'markup', 1)
        treeview.append_column(col)
        selection = treeview.get_selection()
        selection.connect('changed', self.on_selected_datasource_changed)
        treeview.connect('row-activated', self.on_row_activated)

    def refresh_datasources(self):
        treeview = self.builder.get_object('list_datasources')
        model = treeview.get_model()
        model.clear()
        for datasource in self.app.datasources.get_all():
            if datasource.name:
                label = ('<b>%s (%s)</b>'
                         % (gobject.markup_escape_text(datasource.name),
                            gobject.markup_escape_text(datasource.public_url)))
            else:
                label = ('<b>%s</b>'
                         % gobject.markup_escape_text(datasource.public_url))
            if datasource.description:
                label += ('\n<span size="small">%s</span>'
                          % gobject.markup_escape_text(datasource.description))
            else:
                label += ('\n<span size="small"><i>%s</i></span>'
                          % _(u'No description available.'))
            model.append([datasource, label, datasource.color])

    def get_selected_datasource(self):
        treeview = self.builder.get_object('list_datasources')
        selection = treeview.get_selection()
        model, iter_ = selection.get_selected()
        if iter_ is None:
            return None
        return model.get_value(iter_, 0)

    def on_datasource_new(self, *args):
        dlg = DatasourceEditDialog(self.app, self.dlg)
        dlg.run()
        dlg.destroy()

    def on_row_activated(self, treeview, path, column):
        model = treeview.get_model()
        iter_ = model.get_iter(path)
        self.edit_datasource(model.get_value(iter_, 0))

    def on_selected_datasource_changed(self, selection):
        model, iter_ = selection.get_selected()
        for name in ('btn_datasource_edit', 'btn_datasource_delete'):
            widget = self.builder.get_object(name)
            widget.set_sensitive(iter_ is not None)

    def on_edit_datasource(self, *args):
        datasource = self.get_selected_datasource()
        if datasource is None:
            return
        self.edit_datasource(datasource)

    def edit_datasource(self, datasource):
        """Create or edit a data source."""
        dlg = DatasourceEditDialog(self.app, self.dlg, datasource)
        dlg.run()
        dlg.destroy()

    def on_delete_datasource(self, *args):
        datasource = self.get_selected_datasource()
        if datasource is None:
            return
        answer = dialogs.yesno(_(u'Delete Data Source?'), parent=self.dlg)
        if answer == gtk.RESPONSE_YES:
            self.app.datasources.delete(datasource)

    # --------------
    # Provider page
    # --------------

    def _provider_visible(self, model, iter):
        """visible filter function for provider list"""
        combo = self.builder.get_object('combo_provider_filter')
        cmodel = combo.get_model()
        citer = combo.get_active_iter()
        selected = cmodel.get_value(citer, 0)
        available = model.get_value(iter, 4)
        if selected == 0 and not available:  # active
            return False
        elif selected == 1 and available:  # available
            return False
        return True

    def on_provider_filter_changed(self, *args):
        self.refresh_providers()

    def on_refresh_provider(self, *args):
        self.refresh_providers()


    def init_providers(self):
        # List
        treeview = self.builder.get_object('treeview_provider')
        model = gtk.ListStore(str,  # name (= key)
                              str,  # stock icon
                              str,  # label (markup)
                              str,  # human readable name (for sorting)
                              bool, # available
                              )
        model.set_sort_column_id(3, gtk.SORT_ASCENDING)
        filtered_model = model.filter_new()
        filtered_model.set_visible_func(self._provider_visible)
        treeview.set_model(filtered_model)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'stock-id', 1)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'markup', 2)
        treeview.append_column(col)

    def refresh_providers(self):
        available = cf.db.availability()
        model = self.builder.get_object('treeview_provider').get_model()
        model = model.props.child_model
        model.clear()
        for name in cf.db.DIALECTS:
            data = cf.db.DIALECTS[name]
            label = ('<b>%s</b>\n<span size="small">%s</span>'
                     % (gobject.markup_escape_text(data['name']),
                        gobject.markup_escape_text(data['description'])))
            if available[name][0] == True:
                icon = 'gtk-apply'
            else:
                icon = 'gtk-dialog-warning'
                label += (' <span foreground="red" size="small">(%s)</span>'
                          % gobject.markup_escape_text(available[name][1]))
            model.append([name, icon, label, data['name'],
                          available[name][0]])

    def destroy(self):
        while self._ds_sigs:
            self.app.datasources.disconnect(self._ds_sigs.pop())
        self.dlg.destroy()

    def run(self):
        return self.dlg.run()


class DatasourceEditDialog(object):

    def __init__(self, app, parent=None, datasource=None):
        self.app = app
        self.builder = gtk.Builder()
        self.builder.set_translation_domain('crunchyfrog')
        self.builder.add_from_file(app.get_glade_file('datasourceedit.glade'))
        self.builder.connect_signals(self)
        self.dlg = self.builder.get_object('dialog_datasource_edit')
        self.dlg.set_transient_for(parent)
        self.datasource = None
        self.widget_startup_commands = None
        self.populate_dbtype()
        self.setup_startup_commands()
        if datasource:
            self.set_datasource(datasource)

    def run(self):
        return self.dlg.run()

    def destroy(self):
        self.dlg.destroy()

    def on_response(self, dlg, response_id):
        if response_id == 1:  # test connection
            dlg.stop_emission('response')
            self.test_connection()
            return True
        elif response_id == 2:  # ok button
            res = self.save_datasource()
            if not res:
                dlg.stop_emission('response')
                return True

    def populate_dbtype(self):
        model = self.builder.get_object('model_dbtype')
        model.clear()
        model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        available = [key for key, value in cf.db.availability().iteritems()
                     if value[0] == True]
        for key in cf.db.DIALECTS:
            if key not in available:
                continue
            model.append([key, cf.db.DIALECTS[key]['name']])

    def setup_startup_commands(self):
        sw = self.builder.get_object('sw_startup_commands')
        editor = SQLView(self)
        editor.set_show_line_numbers(False)
        sw.add(editor)
        editor.show()
        self.widget_startup_commands = editor

    def set_datasource(self, datasource):
        combo = self.builder.get_object('combo_dbtype')
        model = combo.get_model()
        iter_ = model.get_iter_first()
        while iter_:
            if model.get_value(iter_, 0) == datasource.url.drivername:
                combo.set_active_iter(iter_)
                break
            iter_ = model.iter_next(iter_)
        self.builder.get_object('entry_name').set_text(datasource.name or '')
        entry = self.builder.get_object('entry_description')
        entry.set_text(datasource.description or '')
        backend = self.get_selected_backend()
        backend_options = {}
        [backend_options.setdefault(option.key, option)
         for option in backend.get_options()]
        data = datasource.to_dict()
        for child in self.builder.get_object('table_connection_settings'):
            key = child.get_data('cf::key')
            if key is None:
                continue
            if key in data and key in backend_options:
                option = backend_options[key]
                if option.widget == option.WIDGET_CHECKBOX:
                    child.set_active(data[key])
                elif option.widget == option.WIDGET_FILECHOOSER:
                    child.set_filename(data[key])
                elif option.widget == option.WIDGET_COMBO:
                    model = child.get_model()
                    iter_ = model.get_iter_first()
                    while iter_:
                        if model.get_value(iter_, 0) == data[key]:
                            child.set_active_iter(iter_)
                            break
                        iter_ = model.iter_next(iter_)
                else:
                    if data[key] is None:
                        val = ''
                    else:
                        val = str(data[key])
                    child.set_text(val)
            elif key == 'ask_for_password':
                child.set_active(data['ask_for_password'])
        buffer_ = self.widget_startup_commands.get_buffer()
        buffer_.set_text(datasource.startup_commands or '')
        check = self.builder.get_object('check_color')
        check.set_active(datasource.color is not None)
        if datasource.color is not None:
            btn = self.builder.get_object('colorbutton_datasource')
            btn.set_color(gtk.gdk.color_parse(datasource.color))
        self.datasource = datasource

    def get_selected_backend(self):
        """Returns the selected backend or ``None``."""
        combo = self.builder.get_object('combo_dbtype')
        model = combo.get_model()
        iter_ = combo.get_active_iter()
        if iter_ is None:
            return None
        return cf.db.get_dialect_backend(model.get_value(iter_, 0))

    def on_check_color_toggled(self, check):
        btn = self.builder.get_object('colorbutton_datasource')
        btn.set_sensitive(check.get_active())

    def on_dbtype_changed(self, combo):
        backend = self.get_selected_backend()
        for item in ['lbl_name', 'entry_name',
                     'lbl_description', 'entry_description',
                     'lbl_color', 'check_color',
                     'frame_connection', 'btn_test_connection']:
            self.builder.get_object(item).set_sensitive(backend is not None)
        table = self.builder.get_object('table_connection_settings')
        table.foreach(lambda child: table.remove(child))
        if backend is None:
            return
        row = 0
        for option in backend.get_options():
            row += 1
            second_widget = None
            table.resize(row, 2)
            lbl = gtk.Label('%s:' % option.label)
            lbl.set_alignment(0, 0.5)
            if option.widget == option.WIDGET_CHECKBOX:
                entry = gtk.CheckButton()
            elif option.widget == option.WIDGET_PASSWORD:
                entry = gtk.Entry()
                entry.set_visibility(False)
                second_widget = gtk.CheckButton(_(u'Ask for password'))
                cb = lambda chk, e: e.set_sensitive(not chk.get_active())
                second_widget.connect('toggled', cb, entry)
                second_widget.set_data('cf::key', 'ask_for_password')
            elif option.widget == option.WIDGET_FILECHOOSER:
                entry = gtk.FileChooserButton(_(u'Select file'))
            elif option.widget == option.WIDGET_COMBO:
                model = gtk.ListStore(gobject.TYPE_PYOBJECT, str)
                entry = gtk.ComboBox(model)
                cell = gtk.CellRendererText()
                entry.pack_start(cell, True)
                entry.add_attribute(cell, 'text', 1)
                for choice in option.choices:
                    model.append(choice)
            # TODO: Port widget
            else:
                entry = gtk.Entry()
            if option.tooltip is not None:
                entry.set_tooltip_text(option.tooltip)
            entry.set_data('cf::key', option.key)
            if option.key == 'url':
                entry.set_text('%s://' % backend.drivername)
                entry.set_tooltip_text(_(u'An URL'))

            table.attach(lbl, 0, 1, row-1, row,
                         xoptions=gtk.FILL,
                         yoptions=gtk.FILL)
            table.attach(entry, 1, 2, row-1, row,
                         yoptions=gtk.FILL)
            if second_widget:
                row += 1
                table.attach(second_widget, 1, 2, row-1, row,
                             yoptions=gtk.FILL)
        table.show_all()

    def _get_value_from_widget(self, option):
        """Retrieves the value for an backend option from widget."""
        table = self.builder.get_object('table_connection_settings')
        for child in table.get_children():
            key = child.get_data('cf::key')
            if key != option.key:
                continue
            if option.widget == option.WIDGET_CHECKBOX:
                return child.get_active()
            elif option.widget == option.WIDGET_FILECHOOSER:
                return child.get_filename()
            elif option.widget == option.WIDGET_COMBO:
                model = child.get_model()
                iter_ = child.get_active_iter()
                if iter_ is None:
                    return None
                return model.get_value(iter_, 0)
            else:
                return child.get_text().strip() or None
        return None

    def _get_url_options(self):
        """Collects URL related options."""
        backend = self.get_selected_backend()
        options = {}
        for option in backend.get_options():
            value = self._get_value_from_widget(option)
            options[option.key] = value
            if option.key == 'password':
                opt = GUIOption('ask_for_password', '',
                                widget=GUIOption.WIDGET_CHECKBOX)
                opt.key = 'ask_for_password'
                options['ask_for_password'] = self._get_value_from_widget(opt)
        return options

    def get_sa_url(self):
        """Create and return SQLAlchemy URL from widgets."""
        options = self._get_url_options()
        backend = self.get_selected_backend()
        return backend.create_url(options)

    def make_datasource(self, create_new=False):
        """Apply current settings to data source or create a new one."""
        if self.datasource and not create_new:
            ds = self.datasource
        else:
            ds = cf.db.Datasource(self.app.datasources)
        ds.url = self.get_sa_url()
        data = self._get_url_options()
        if 'ask_for_password' in data:
            ds.ask_for_password = data['ask_for_password']
        ds.name = self.builder.get_object('entry_name').get_text() or None
        widget = self.builder.get_object('entry_description')
        ds.description = widget.get_text() or None
        if self.builder.get_object('check_color').get_active():
            btn = self.builder.get_object('colorbutton_datasource')
            ds.color = btn.get_color().to_string()
        else:
            ds.color = None
        buffer_ = self.widget_startup_commands.get_buffer()
        ds.startup_commands = buffer_.get_text(*buffer_.get_bounds()) or None
        return ds

    def clean_data(self):
        """Validates the form data."""
        data = self._get_url_options()
        backend = self.get_selected_backend()
        required = []
        errors = []
        for option in backend.get_options():
            value = data.get(option.key, None)
            if option.required and value is None:
                required.append(option)
            elif value and option.widget == option.WIDGET_PORT:
                try:
                    int(value)
                except (ValueError, TypeError):
                    errors.append((option, _(u'Integer required')))
        return {'required': required, 'errors': errors}

    def validate_data(self):
        """Validates data and displays dialog on errors."""
        result = self.clean_data()
        msg = ''
        if result['required']:
            msg += _(u'Missing fields:\n%(fields)s')
            msg = msg % {'fields': ', '.join(
                [opt.label for opt in result['required']])}
        if result['errors']:
            msg += _(u'Validation errors:\n')
            msg += '\n'.join('%s: %s' % (opt.label, err)
                             for opt, err in result['errors'])
        if msg:
            dialogs.error(_(u'Failed'), msg, parent=self.dlg)
            return False
        return True

    def test_connection(self):
        """Test the current settings."""
        if not self.validate_data():
            return
        ds = self.make_datasource(create_new=True)
        try:
            conn = ds._open_connection()
            conn.close()
            dialogs.info(_(u'Succeeded'), parent=self.dlg)
        except:
            dialogs.error(_(u'Failed'), str(sys.exc_info()[1]))

    def save_datasource(self):
        """Save current settings."""
        if not self.validate_data():
            return False
        ds = self.make_datasource()
        self.app.datasources.save(ds)
        return True



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Informix backend."""

from cf.db.backends import Generic, DEFAULT_OPTIONS


class Informix(Generic):

    drivername = 'informix'

    @classmethod
    def dbapi(cls):
        import informixdb
        return informixdb

    @classmethod
    def get_connect_params(cls, url):
        # copied from sqlalchemy.databases.informix
        if url.host:
            dsn = '%s@%s' % ( url.database , url.host )
        else:
            dsn = url.database

        if url.username:
            opt = { 'user':url.username , 'password': url.password }
        else:
            opt = {}

        return ([dsn], opt)

    @classmethod
    def get_options(cls):
        return DEFAULT_OPTIONS


DRIVER = Informix
